# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


import pytest
from random import randrange

from aiootp.asynchs import ConcurrencyGuard

from conftest import *


class TestOnlineCipherInterfaces:
    async def test_sync_cipher_decipher_streams_with_varied_data_sizes(
        self,
    ) -> None:
        for config, cipher, salt, aad in all_ciphers:
            for i in range(
                0, 512 - config.INNER_HEADER_BYTES - 1, randrange(1, 64)
            ):
                stream_enc = cipher.stream_encrypt(salt=salt, aad=aad)
                pt_enc = i * b"a"
                stream_enc.buffer(pt_enc)

                stream_dec = cipher.stream_decrypt(
                    salt=stream_enc.salt,
                    aad=stream_enc.aad,
                    iv=stream_enc.iv,
                )
                for var in ("salt", "iv", "aad", "PACKETSIZE"):
                    assert getattr(stream_enc, var) == getattr(
                        stream_dec, var
                    )
                pt_dec = b""
                join = b"".join
                for id_ct in stream_enc:
                    problem = (  # fmt: skip
                        "Invalid packet sizes were allowed."
                    )
                    with Ignore(ValueError, if_else=violation(problem)):
                        stream_dec.buffer(
                            join(id_ct)[: stream_dec.PACKETSIZE - 1]
                        )

                    stream_dec.buffer(join(id_ct))
                    for pt in stream_dec:
                        pt_dec += pt
                for id_ct in stream_enc.finalize():
                    stream_dec.buffer(join(id_ct))
                for pt in stream_dec.finalize():
                    pt_dec += pt

                problem = (  # fmt: skip
                    "Processing was allowed to continue after finalization."
                )
                with Ignore(InterruptedError, if_else=violation(problem)):
                    stream_enc.buffer(pt_enc)
                with Ignore(InterruptedError, if_else=violation(problem)):
                    stream_dec.buffer(b"".join(id_ct))
                stream_dec.shmac.test_shmac(stream_enc.shmac.result)
                assert (
                    pt_enc == pt_dec
                ), f"{i=} : plaintext_len={len(pt_enc)} : decrypted_plaintext_len={len(pt_dec)}"

    async def test_async_cipher_decipher_streams_with_varied_data_sizes(
        self,
    ) -> None:
        for config, cipher, salt, aad in all_ciphers:
            for i in range(
                0, 512 - config.INNER_HEADER_BYTES - 1, randrange(1, 64)
            ):
                stream_enc = await cipher.astream_encrypt(
                    salt=salt, aad=aad
                )
                pt_enc = i * b"a"
                await stream_enc.abuffer(pt_enc)

                stream_dec = await cipher.astream_decrypt(
                    salt=stream_enc.salt,
                    aad=stream_enc.aad,
                    iv=stream_enc.iv,
                )
                for var in ("salt", "iv", "aad", "PACKETSIZE"):
                    assert getattr(stream_enc, var) == getattr(
                        stream_dec, var
                    )
                pt_dec = b""
                join = b"".join
                async for id_ct in stream_enc:
                    problem = (  # fmt: skip
                        "Invalid packet sizes were allowed."
                    )
                    with Ignore(ValueError, if_else=violation(problem)):
                        await stream_dec.abuffer(
                            join(id_ct)[: stream_dec.PACKETSIZE - 1]
                        )

                    await stream_dec.abuffer(join(id_ct))
                    async for pt in stream_dec:
                        pt_dec += pt
                async for id_ct in stream_enc.afinalize():
                    await stream_dec.abuffer(join(id_ct))
                async for pt in stream_dec.afinalize():
                    pt_dec += pt

                problem = (  # fmt: skip
                    "Processing was allowed to continue after finalization."
                )
                with Ignore(InterruptedError, if_else=violation(problem)):
                    await stream_enc.abuffer(pt_enc)
                with Ignore(InterruptedError, if_else=violation(problem)):
                    await stream_dec.abuffer(b"".join(id_ct))
                await stream_dec.shmac.atest_shmac(stream_enc.shmac.result)
                assert (
                    pt_enc == pt_dec
                ), f"{i=} : plaintext_len={len(pt_enc)} : decrypted_plaintext_len={len(pt_dec)}"

    async def test_async_encipher_sync_decipher_interop(self) -> None:
        for _, cipher, salt, aad in all_ciphers:
            stream_enc = await cipher.astream_encrypt(salt=salt, aad=aad)
            pt_enc = plaintext_bytes
            await stream_enc.abuffer(pt_enc)

            stream_dec = cipher.stream_decrypt(
                salt=stream_enc.salt, aad=stream_enc.aad, iv=stream_enc.iv
            )
            for var in ("salt", "iv", "aad", "PACKETSIZE"):
                assert getattr(stream_enc, var) == getattr(stream_dec, var)
            pt_dec = b""
            join = b"".join

            async for block_id, block in stream_enc:
                # the cipher can be resumed after the failure state of an
                # InvalidBlockID exception is recovered & the correct data
                # is supplied to the buffer
                problem = (  # fmt: skip
                    "An altered block_id was not detected."
                )
                fake_block_id = xi_mix(
                    block_id + b"\x01", size=len(block_id)
                )
                with Ignore(
                    InvalidBlockID, if_else=violation(problem)
                ) as relay:
                    stream_dec.buffer(fake_block_id + block)
                assert fake_block_id == relay.error.failure_state.block_id
                assert block == relay.error.failure_state.block
                assert 0 == len(relay.error.failure_state.buffer())

                problem = (  # fmt: skip
                    "An altered block was not detected."
                )
                fake_block = xi_mix(block + b"\x01", size=len(block))
                with Ignore(
                    InvalidBlockID, if_else=violation(problem)
                ) as relay:
                    stream_dec.buffer(block_id + fake_block)
                assert block_id == relay.error.failure_state.block_id
                assert fake_block == relay.error.failure_state.block
                assert 0 == len(relay.error.failure_state.buffer())

                # correct data is processed without failure
                stream_dec.buffer(block_id + block)
                for pt in stream_dec:
                    pt_dec += pt
            async for id_ct in stream_enc.afinalize():
                stream_dec.buffer(join(id_ct))
            for pt in stream_dec.finalize():
                pt_dec += pt
            await stream_dec.shmac.atest_shmac(stream_enc.shmac.result)
            assert pt_dec == pt_enc

    async def test_sync_encipher_async_decipher_interop(self) -> None:
        for _, cipher, salt, aad in all_ciphers:
            stream_enc = cipher.stream_encrypt(salt=salt, aad=aad)
            pt_enc = plaintext_bytes
            stream_enc.buffer(pt_enc)

            stream_dec = await cipher.astream_decrypt(
                salt=stream_enc.salt, aad=stream_enc.aad, iv=stream_enc.iv
            )
            for var in ("salt", "iv", "aad", "PACKETSIZE"):
                assert getattr(stream_enc, var) == getattr(stream_dec, var)
            pt_dec = b""
            join = b"".join

            for block_id, block in stream_enc:
                # the cipher can be resumed after the failure state of an
                # InvalidBlockID exception is recovered & the correct data
                # is supplied to the buffer
                problem = (  # fmt: skip
                    "An altered block_id was not detected."
                )
                fake_block_id = xi_mix(
                    block_id + b"\x01", size=len(block_id)
                )
                async with Ignore(
                    InvalidBlockID, if_else=violation(problem)
                ) as relay:
                    await stream_dec.abuffer(fake_block_id + block)
                assert fake_block_id == relay.error.failure_state.block_id
                assert block == relay.error.failure_state.block
                assert 0 == len(relay.error.failure_state.buffer())

                problem = (  # fmt: skip
                    "An altered block was not detected."
                )
                fake_block = xi_mix(block + b"\x01", size=len(block))
                async with Ignore(
                    InvalidBlockID, if_else=violation(problem)
                ) as relay:
                    await stream_dec.abuffer(block_id + fake_block)
                assert block_id == relay.error.failure_state.block_id
                assert fake_block == relay.error.failure_state.block
                assert 0 == len(relay.error.failure_state.buffer())

                # correct data is processed without failure
                await stream_dec.abuffer(block_id + block)
                async for pt in stream_dec:
                    pt_dec += pt

            for id_ct in stream_enc.finalize():
                await stream_dec.abuffer(join(id_ct))
            async for pt in stream_dec.afinalize():
                pt_dec += pt
            stream_dec.shmac.test_shmac(stream_enc.shmac.result)
            assert pt_dec == pt_enc

    async def test_data_type_must_be_bytes(self) -> None:
        problem = (  # fmt: skip
            "A non-bytes data type was allowed to be buffered."
        )
        for config, cipher, salt, _ in all_ciphers:
            for data in ("test", 0, None):
                with Ignore(TypeError, if_else=violation(problem)):
                    astream_enc = await cipher.astream_encrypt()
                    await astream_enc.abuffer(data)

                with Ignore(TypeError, if_else=violation(problem)):
                    stream_enc = cipher.stream_encrypt()
                    stream_enc.buffer(data)

                errors = (TypeError, ValueError)

                with Ignore(*errors, if_else=violation(problem)):
                    astream_dec = await cipher.astream_decrypt(
                        salt=salt, iv=csprng(config.IV_BYTES)
                    )
                    await astream_dec.abuffer(data)

                with Ignore(*errors, if_else=violation(problem)):
                    stream_dec = cipher.stream_decrypt(
                        salt=salt, iv=csprng(config.IV_BYTES)
                    )
                    stream_dec.buffer(data)

    async def test_async_buffer_concurrency_handling(self) -> None:
        config, cipher, salt, aad = choice(all_ciphers)
        chunk_size = 1024 * config.BLOCKSIZE
        data_a = (chunk_size * b"a")[config.INNER_HEADER_BYTES :]
        data_b = (chunk_size * b"b")[: -config.SENTINEL_BYTES]

        stream_enc = await cipher.astream_encrypt(salt=salt, aad=aad)
        fut_a = asynchs.new_task(stream_enc.abuffer(data_a))
        await asleep(0.00001)
        fut_b = asynchs.new_task(stream_enc.abuffer(data_b))
        await fut_a
        await fut_b
        ct = [b"".join(id_ct) async for id_ct in stream_enc.afinalize()]

        stream_dec = await cipher.astream_decrypt(
            salt=salt, aad=aad, iv=stream_enc.iv
        )
        ct_a = b"".join(ct[: len(ct) // 2])
        ct_b = b"".join(ct[len(ct) // 2 :])
        fut_a = asynchs.new_task(stream_dec.abuffer(ct_a))
        await asleep(0.00001)
        fut_b = asynchs.new_task(stream_dec.abuffer(ct_b))
        await fut_a
        await fut_b
        pt = b"".join([pt async for pt in stream_dec.afinalize()])

        assert data_a + data_b == pt

    async def test_sync_buffer_concurrency_handling(self) -> None:
        config, cipher, salt, aad = choice(all_ciphers)
        chunk_size = 16 * 1024 * config.BLOCKSIZE
        data_a = (chunk_size * b"a")[config.INNER_HEADER_BYTES :]
        data_b = (chunk_size * b"b")[: -config.SENTINEL_BYTES]

        stream_enc = cipher.stream_encrypt(salt=salt, aad=aad)
        fut_a = Threads.submit(stream_enc.buffer, data_a)
        asynchs.sleep(0.005)
        fut_b = Threads.submit(stream_enc.buffer, data_b)
        fut_a.result()
        fut_b.result()
        ct = [b"".join(id_ct) for id_ct in stream_enc.finalize()]

        stream_dec = cipher.stream_decrypt(
            salt=salt, aad=aad, iv=stream_enc.iv
        )
        ct_a = b"".join(ct[: len(ct) // 2])
        ct_b = b"".join(ct[len(ct) // 2 :])
        fut_a = Threads.submit(stream_dec.buffer, ct_a)
        asynchs.sleep(0.005)
        fut_b = Threads.submit(stream_dec.buffer, ct_b)
        fut_a.result()
        fut_b.result()
        pt = b"".join(stream_dec.finalize())

        assert data_a + data_b == pt

    async def test_async_finalize_concurrency_handling(self) -> None:
        config, cipher, salt, aad = choice(all_ciphers)
        chunk_size = 1024 * config.BLOCKSIZE
        data = chunk_size * b"a"

        stream_enc = await cipher.astream_encrypt(salt=salt, aad=aad)
        await stream_enc.abuffer(data)
        finalizing = stream_enc.afinalize()
        ct = b"".join(await finalizing.asend(None))

        problem = (  # fmt: skip
            "Multiple calls to afinalize were allowed."
        )
        with Ignore(
            ConcurrencyGuard.IncoherentConcurrencyState,
            if_else=violation(problem),
        ):
            async for _ in stream_enc.afinalize():
                pytest.fail(problem)

        ct += b"".join([b"".join(id_ct) async for id_ct in finalizing])

        stream_dec = await cipher.astream_decrypt(
            salt=salt, aad=aad, iv=stream_enc.iv
        )
        await stream_dec.abuffer(ct)
        finalizing = stream_dec.afinalize()
        pt = await finalizing.asend(None)

        problem = (  # fmt: skip
            "Multiple calls to afinalize were allowed."
        )
        with Ignore(
            ConcurrencyGuard.IncoherentConcurrencyState,
            if_else=violation(problem),
        ):
            async for _ in stream_dec.afinalize():
                pytest.fail(problem)

        pt += b"".join([pt async for pt in finalizing])
        assert data == pt

    async def test_sync_finalize_concurrency_handling(self) -> None:
        config, cipher, salt, aad = choice(all_ciphers)
        chunk_size = 1024 * config.BLOCKSIZE
        data = chunk_size * b"a"

        stream_enc = cipher.stream_encrypt(salt=salt, aad=aad).buffer(data)
        finalizing = stream_enc.finalize()
        ct = b"".join(finalizing.send(None))

        problem = (  # fmt: skip
            "Multiple calls to finalize were allowed."
        )
        with Ignore(
            ConcurrencyGuard.IncoherentConcurrencyState,
            if_else=violation(problem),
        ):
            for _ in stream_enc.finalize():
                pytest.fail(problem)

        ct += b"".join(b"".join(id_ct) for id_ct in finalizing)

        stream_dec = cipher.stream_decrypt(
            salt=salt, aad=aad, iv=stream_enc.iv
        ).buffer(ct)
        finalizing = stream_dec.finalize()
        pt = finalizing.send(None)

        problem = (  # fmt: skip
            "Multiple calls to finalize were allowed."
        )
        with Ignore(
            ConcurrencyGuard.IncoherentConcurrencyState,
            if_else=violation(problem),
        ):
            for _ in stream_dec.finalize():
                pytest.fail(problem)

        pt += b"".join(finalizing)
        assert data == pt


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
An interface for feeding & retrieving from a SHA3 XOF object used
as an entropy pool.
"""

__all__ = ["ThreadingSafeEntropyPool"]


from secrets import token_bytes

from aiootp._typing import Typing as t
from aiootp._constants import INT_BYTES, BIG
from aiootp.commons import FrozenInstance
from aiootp.generics import acanonical_pack, canonical_pack

from .simple import acanonical_token, canonical_token


class ThreadingSafeEntropyPool(FrozenInstance):
    """
    A definition for bidirectional sponge objects which incorporate,
    store, & extract useful entropy.
    """

    __slots__ = ("_obj", "_pool")

    def __init__(
        self,
        data: bytes,
        *,
        obj: t.Callable[[bytes], t.Union[t.HasherType, t.XOFType]],
        pool: t.Sequence[bytes],
    ) -> None:
        """
        Copies over the object dictionary of the `obj` hashing object.
        """
        self._pool = pool
        self._obj = obj(data)
        self._obj.update(token_bytes(3 * self._obj.block_size))

    @property
    def name(self) -> str:
        return self._obj.name

    @property
    def block_size(self) -> int:
        return self._obj.block_size

    @property
    def digest_size(self) -> int:
        return self._obj.digest_size

    @property
    def update(self) -> t.Callable[[bytes], None]:
        return self._obj.update

    @property
    def digest(self) -> t.Callable[..., bytes]:
        """
        ************************
        BEWARE: NOT THREAD SAFE.
        ************************
        """
        return self._obj.digest

    @property
    def hexdigest(self) -> t.Callable[..., bytes]:
        """
        ************************
        BEWARE: NOT THREAD SAFE.
        ************************
        """
        return self._obj.hexdigest

    def copy(self) -> t.Cls:
        """
        Allows the user to create a copy instance of the hashing object.
        """
        new_self = self.__class__.__new__(self.__class__)
        new_self._obj = self._obj.copy()
        new_self._pool = self._pool.copy()
        return new_self

    async def _amake_token(self, *data: bytes) -> bytes:
        """
        Returns a PRIVATE, unique, & canonically encoded token.
        """
        return await acanonical_pack(
            await acanonical_token(),
            self._pool[0],
            *data,
            blocksize=self.block_size,
            pad=b"\x5c",
        )

    def _make_token(self, *data: bytes) -> bytes:
        """
        Returns a PRIVATE, unique, & canonically encoded token.
        """
        return canonical_pack(
            canonical_token(),
            self._pool[0],
            *data,
            blocksize=self.block_size,
            pad=b"\x36",
        )

    async def ahash(self, *data: bytes, size: int) -> bytes:
        """
        Receives any number of arguments of bytes type `data` &
        updates the instance with them all sequentially & canonically
        encoded.
        """
        digest_size = size.to_bytes(INT_BYTES, BIG)
        token = await self._amake_token(digest_size, *data)
        self.update(token)
        entropy = self._obj.copy()
        entropy.update(token)
        return entropy.digest(size) if size else entropy.digest()

    def hash(self, *data: bytes, size: int) -> bytes:
        """
        Receives any number of arguments of bytes type `data` &
        updates the instance with them all sequentially & canonically
        encoded.
        """
        digest_size = size.to_bytes(INT_BYTES, BIG)
        token = self._make_token(digest_size, *data)
        self.update(token)
        entropy = self._obj.copy()
        entropy.update(token)
        return entropy.digest(size) if size else entropy.digest()


module_api = dict(
    ThreadingSafeEntropyPool=t.add_type(ThreadingSafeEntropyPool),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Types to manage bundled cipher session values.
"""

__all__ = ["KeyAADBundle", "KeyAADMode", "KeyAADRegisters", "SaltAADIV"]


from secrets import token_bytes

from aiootp._typing import Typing as t
from aiootp._constants import DEFAULT_AAD, ASYNC, SYNC
from aiootp._exceptions import Issue, KeyAADIssue
from aiootp.commons import FrozenSlots, OpenFrozenSlots
from aiootp.generics import canonical_pack

from .cipher_kdfs import CipherKDFs


class SaltAADIV(FrozenSlots):
    """
    Creates efficient containers for salt, aad, IV value bundles.
    """

    __slots__ = ("salt", "aad", "iv", "config", "iv_is_fresh")

    _MAPPED_ATTRIBUTES: t.Tuple[str] = ("salt", "aad", "iv")

    def __init__(
        self,
        salt: t.Optional[bytes] = None,
        aad: bytes = DEFAULT_AAD,
        iv: t.Optional[bytes] = None,
        *,
        config: t.ConfigType,
    ) -> None:
        self.config = config
        self.salt = salt if salt else token_bytes(self.config.SALT_BYTES)
        self.aad = aad
        self.iv = self._process_iv(iv)
        self._test_salt_aad_iv(self.salt, self.aad, self.iv)

    def __iter__(self) -> t.Generator[str, None, None]:
        yield from self._MAPPED_ATTRIBUTES

    def _process_iv(self, iv: bytes) -> bytes:
        """
        Creating a new random IV during encryption ensures the user
        cannot accidentally prevent the derived key material from being
        fresh. A flag allows improper usage to be acted upon downstream.
        """
        if iv is None:
            self.iv_is_fresh = True
            return token_bytes(self.config.IV_BYTES)
        else:
            self.iv_is_fresh = False
            return iv

    def _test_salt_aad_iv(self, salt: bytes, aad: bytes, iv: bytes) -> None:
        """
        Validates the ephemeral `salt`, `aad` authenticated associated
        data, & the random `iv` for a package cipher.
        """
        config = self.config
        if salt.__class__ is not bytes:
            raise Issue.value_must_be_type("salt", bytes)
        elif len(salt) != config.SALT_BYTES:
            raise Issue.invalid_length("salt", config.SALT_BYTES)
        elif aad.__class__ is not bytes:
            raise Issue.value_must_be_type("aad", bytes)
        elif iv.__class__ is not bytes:
            raise Issue.value_must_be_type("iv", bytes)
        elif len(iv) != config.IV_BYTES:
            raise Issue.invalid_length("iv", config.IV_BYTES)


class KeyAADRegisters(FrozenSlots):
    """
    Efficiently stores objects which help to enforce the limited usage
    of a `KeyAADBundle` object for a single encryption / decryption
    round.
    """

    __slots__ = ("keystream", "shmac")

    def register(self, name: str, value: t.Any) -> None:
        setattr(self, name, value)


class KeyAADMode(OpenFrozenSlots):
    """
    Helps guide users towards correct usage of `KeyAADBundle` objects in
    ciphers by enforcing that they are set to async or sync key
    derivation modes when using them in those contexts.
    """

    __slots__ = ("_mode",)

    def __eq__(self, mode: str) -> bool:
        """
        The object can directly be compared to the string `mode` from
        within the runtime of an async & sync contexts. Procs an error
        if the mode has not been set.
        """
        return self.mode == mode

    @property
    def mode(self) -> str:
        """
        Procs an error if the mode has not been set.
        """
        try:
            return self._mode
        except AttributeError as error:
            raise KeyAADIssue.KDFModeNotDeclared from error

    def set_async_mode(self) -> None:
        """
        Sets the object's mode to signal async key derivation is needed.
        """
        self._mode = ASYNC

    def set_sync_mode(self) -> None:
        """
        Sets the object's mode to signal sync key derivation is needed.
        """
        self._mode = SYNC

    def validate(self) -> None:
        """
        Procs an error if the mode has not been set, else returns `None`.
        """
        return self.mode and None


class KeyAADBundle:
    """
    A low-level interface for managing a key, salt, iv & authenticated
    associated data bundle which is to be used for ONLY ONE encryption.
    """

    __slots__ = (
        "_kdfs",
        "_session",
        "_bundle",
        "_mode",
        "_registers",
        "config",
    )

    _Session: type
    _Mode: type = KeyAADMode
    _Registers: type = KeyAADRegisters
    _SaltAADIV: type = SaltAADIV

    def __init__(
        self,
        kdfs: CipherKDFs,
        *,
        salt: t.Optional[bytes] = None,
        aad: bytes = DEFAULT_AAD,
        iv: t.Optional[bytes] = None,
    ) -> None:
        """
        Stores the `salt`, `aad` & `iv` & initializes the session KDFs
        for this permutation of the values.

        The `iv` should only be passed when the bundle is to be used for
        decryption.
        """
        if not issubclass(kdfs.__class__, CipherKDFs):
            raise Issue.value_must_be_type("kdfs", CipherKDFs)
        self.config = kdfs.config
        self._kdfs = kdfs
        self._session = self._Session()
        self._mode = self._Mode()
        self._registers = self._Registers()
        self._bundle = self._SaltAADIV(
            salt=salt,
            aad=aad,
            iv=iv,
            config=self.config,
        )
        self._initialize_session()

    def __iter__(self) -> t.Iterable[bytes]:
        """
        Yields the instance's salt, authenticated associated data, &
        the IV.
        """
        yield self._bundle.salt
        yield self._bundle.aad
        yield self._bundle.iv

    def _initialize_session(self) -> None:
        """
        Initializes cipher KDFs with a canonicalized session summary.
        """
        session = self._session
        summary = canonical_pack(*self, int_bytes=4)
        for kdf_name, kdf_session_copy in self._kdfs.new_session(summary):
            setattr(session, kdf_name, kdf_session_copy)

    def _register_shmac(self, shmac: t.StreamHMACType) -> None:
        """
        Registers the shmac which will be tied to the instance for a
        single run of the cipher. Reusing an instance for multiple
        cipher calls is NOT SAFE, & is disallowed by this registration.
        """
        if hasattr(self._registers, "shmac"):
            raise KeyAADIssue.shmac_already_registered()
        self._registers.register("shmac", shmac)

    @property
    def salt(self) -> bytes:
        """
        A [pseudo]random salt that may be supplied by the user. By
        default it's sent in the clear attached to the ciphertext.
        Thus it may simplify implementing efficient features, such
        as search or routing, though care must still be taken when
        considering how leaking such metadata may be harmful.

        Keeping this value constant is strongly discouraged. Though,
        the cipher's salt misuse-reuse resistance is ruled by the
        combination of the automatically incorporated `timestamp`,
        `iv`, & `siv_key`. The risk calculation starts with setting
        r = len(iv + siv_key) / 3. Then, all else staying constant,
        once 256**r messages are encrypted within a second, each
        additional encrypted message within that same second begins
        to have more than a 256**(-r) chance of generating a repeat
        context.

        See: https://github.com/rmlibre/aiootp/issues/16
        """
        return self._bundle.salt

    @property
    def aad(self) -> bytes:
        """
        An arbitrary bytes value that a user decides to categorize
        keystreams. It is authenticated as associated data & safely
        differentiates keystreams as a tweak when it's unique for
        each permutation of `key`, `salt`, & `iv`.
        """
        return self._bundle.aad

    @property
    def _iv_given_by_user(self) -> bool:
        """
        Returns a boolean flag to determine inappropriate usage of the
        IV.
        """
        return not self._bundle.iv_is_fresh

    @property
    def iv(self) -> bytes:
        """
        An ephemeral, uniform, random value that's generated by
        the encryption algorithm.
        """
        return self._bundle.iv

    @property
    def _shmac_kdf(self) -> t.XOFType:
        """
        Returns the XOF object used by the `StreamHMAC` class.
        """
        return self._session.shmac_kdf


module_api = dict(
    KeyAADBundle=t.add_type(KeyAADBundle),
    KeyAADMode=t.add_type(KeyAADMode),
    KeyAADRegisters=t.add_type(KeyAADRegisters),
    SaltAADIV=t.add_type(SaltAADIV),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *

from aiootp._gentools import abatch, batch
from aiootp.ciphers import Ciphertext
from aiootp.ciphers.key_bundle import SaltAADIV


plaintext_bytes = token_bytes(1024)


async def aempty_stream() -> t.AsyncGenerator[None, None]:
    yield


def empty_stream() -> t.Generator[None, None, None]:
    yield


class TestDatastreamLimits:
    async def test_async_datastream_must_emit_in_blocksize_chunks(
        self,
    ) -> None:
        for config, cipher, salt, aad in dual_output_ciphers:
            BLOCKSIZE = config.BLOCKSIZE
            ainvalid_size_datastream = abatch(
                plaintext_bytes, size=BLOCKSIZE + 1
            )
            akey_bundle = await cipher._KeyAADBundle(
                cipher._kdfs, salt=salt, aad=aad
            ).async_mode()
            shmac = cipher._StreamHMAC(akey_bundle)._for_encryption()

            problem = (  # fmt: skip
                f"An async datastream block was allowed to exceed {BLOCKSIZE} bytes."
            )
            async with Ignore(OverflowError, if_else=violation(problem)):
                async for _ in cipher._Junction.abytes_encipher(
                    ainvalid_size_datastream, shmac=shmac
                ):
                    pass

    async def test_sync_datastream_must_emit_in_blocksize_chunks(
        self,
    ) -> None:
        for config, cipher, salt, aad in dual_output_ciphers:
            BLOCKSIZE = config.BLOCKSIZE
            invalid_size_datastream = batch(
                plaintext_bytes, size=BLOCKSIZE + 1
            )
            key_bundle = cipher._KeyAADBundle(
                cipher._kdfs, salt=salt, aad=aad
            ).sync_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_encryption()

            problem = (  # fmt: skip
                f"A sync datastream block was allowed to exceed {BLOCKSIZE} bytes."
            )
            with Ignore(OverflowError, if_else=violation(problem)):
                for _ in cipher._Junction.bytes_encipher(
                    invalid_size_datastream, shmac=shmac
                ):
                    pass

    async def test_async_datastream_must_emit_at_least_one_block(
        self,
    ) -> None:
        for _, cipher, salt, aad in dual_output_ciphers:
            ainvalid_size_datastream = aempty_stream()
            await ainvalid_size_datastream.asend(None)
            akey_bundle = await cipher._KeyAADBundle(
                cipher._kdfs, salt=salt, aad=aad
            ).async_mode()
            shmac = cipher._StreamHMAC(akey_bundle)._for_encryption()

            problem = (  # fmt: skip
                "An empty async datastream was allowed."
            )
            async with Ignore(ValueError, if_else=violation(problem)):
                async for _ in cipher._Junction.abytes_encipher(
                    ainvalid_size_datastream, shmac=shmac
                ):
                    pass

    async def test_sync_datastream_must_emit_at_least_one_block(
        self,
    ) -> None:
        for _, cipher, salt, aad in dual_output_ciphers:
            invalid_size_datastream = empty_stream()
            invalid_size_datastream.send(None)
            key_bundle = cipher._KeyAADBundle(
                cipher._kdfs, salt=salt, aad=aad
            ).sync_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_encryption()

            problem = (  # fmt: skip
                "An empty sync datastream was allowed."
            )
            with Ignore(ValueError, if_else=violation(problem)):
                for _ in cipher._Junction.bytes_encipher(
                    invalid_size_datastream, shmac=shmac
                ):
                    pass

    async def test_async_too_large_plaintext_block_overflows_validated_transform(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "An async plaintext block larger than blocksize didn't "
            "overflow."
        )
        for config, cipher, salt, aad in dual_output_ciphers:
            BLOCKSIZE = config.BLOCKSIZE
            # fmt: off
            for mode, kw in (
                ("_for_encryption", dict(salt=salt, aad=aad)),
                ("_for_decryption", dict(salt=salt, aad=aad, iv=token_bytes(config.IV_BYTES))),
            ):
                # fmt: off
                akey_bundle = await cipher._KeyAADBundle(cipher._kdfs, **kw).async_mode()
                shmac = getattr(cipher._StreamHMAC(akey_bundle), mode)()

                with Ignore(OverflowError, if_else=violation(problem)):
                    await shmac._avalidated_transform(
                        b"\xff" + token_bytes(BLOCKSIZE),
                        token_bytes(BLOCKSIZE),
                    )

    async def test_sync_too_large_plaintext_block_overflows_validated_transform(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "A sync plaintext block larger than blocksize didn't "
            "overflow."
        )
        for config, cipher, salt, aad in dual_output_ciphers:
            BLOCKSIZE = config.BLOCKSIZE
            # fmt: off
            for mode, kw in (
                ("_for_encryption", dict(salt=salt, aad=aad)),
                ("_for_decryption", dict(salt=salt, aad=aad, iv=token_bytes(config.IV_BYTES))),
            ):
                # fmt: on
                key_bundle = cipher._KeyAADBundle(cipher._kdfs, **kw).sync_mode()
                shmac = getattr(cipher._StreamHMAC(key_bundle), mode)()

                with Ignore(OverflowError, if_else=violation(problem)):
                    shmac._validated_transform(
                        b"\xff" + token_bytes(BLOCKSIZE),
                        token_bytes(BLOCKSIZE),
                    )


class TestCipherInputs:
    async def test_input_container(self) -> None:
        for config, _, salt, aad in all_ciphers:
            iv = token_bytes(config.IV_BYTES)
            sav = SaltAADIV(salt=salt, aad=aad, iv=iv, config=config)
            container = Namespace({name: sav[name] for name in sav})
            assert container.salt == sav.salt
            assert sav.salt == salt
            assert container.aad == sav.aad
            assert sav.aad == aad
            assert container.iv == sav.iv
            assert sav.iv == iv

    async def test_kdfs_limits(self) -> None:
        class InvalidCipherKDFs:
            pass

        problem = (  # fmt: skip
            "An invalid cipher KDFs type was allowed."
        )
        for _, cipher, *_ in all_ciphers:
            kdfs = InvalidCipherKDFs()
            with Ignore(TypeError, if_else=violation(problem)):
                cipher._KeyAADBundle(kdfs)

    async def test_keys_limits(self) -> None:
        for _, cipher, *_ in all_ciphers:
            problem = (  # fmt: skip
                "Non-bytes key was allowed."
            )
            with Ignore(TypeError, if_else=violation(problem)):
                cipher.__class__(csprng().hex())

            problem = (  # fmt: skip
                "A shorter than min length key was allowed."
            )
            with Ignore(ValueError, if_else=violation(problem)):
                cipher.__class__(csprng(MIN_KEY_BYTES - 1))

    async def test_salt_limits(self) -> None:
        for config, cipher, *_ in all_ciphers:
            problem = (  # fmt: skip
                "Non-bytes salt was allowed."
            )
            with Ignore(TypeError, if_else=violation(problem)):
                cipher._KeyAADBundle(
                    cipher._kdfs, salt=csprng(config.SALT_BYTES // 2).hex()
                )

            problem = (  # fmt: skip
                "Invalid length salt was allowed."
            )
            with Ignore(ValueError, if_else=violation(problem)):
                cipher._KeyAADBundle(cipher._kdfs, salt=csprng())

    async def test_aad_limits(self) -> None:
        for _, cipher, salt, aad in all_ciphers:
            problem = (  # fmt: skip
                "Non-bytes aad was allowed."
            )
            with Ignore(TypeError, if_else=violation(problem)):
                cipher._KeyAADBundle(cipher._kdfs, salt=salt, aad=aad.hex())

    async def test_iv_limits(self) -> None:
        for config, cipher, *_ in all_ciphers:
            problem = (  # fmt: skip
                "Non-bytes IV was allowed."
            )
            with Ignore(TypeError, if_else=violation(problem)):
                cipher._KeyAADBundle(
                    cipher._kdfs, iv=token_bytes(config.IV_BYTES // 2).hex()
                )

            problem = (  # fmt: skip
                "Invalid length IV was allowed."
            )
            with Ignore(ValueError, if_else=violation(problem)):
                cipher._KeyAADBundle(
                    cipher._kdfs, iv=token_bytes(config.IV_BYTES + 1)
                )


class TestSaltMisuseReuseResistance:
    number_of_tests: int = 256

    async def test_async_siv_only_resistance(self) -> None:
        for config, cipher, salt, _ in all_ciphers:
            iv = token_bytes(config.IV_BYTES)
            kw = dict(kdfs=cipher._kdfs, salt=salt, iv=iv)

            # aggregate the first ciphertext block of a collection of async
            # ciphertexts instantiated with the same key, salt, aad & iv
            aciphertexts = set()
            for _ in range(self.number_of_tests):
                plaintext = await cipher._padding.astart_padding() + (
                    (config.BLOCKSIZE - config.INNER_HEADER_BYTES) * b"\x00"
                )
                key_bundle = await cipher._KeyAADBundle(**kw).async_mode()
                object.__setattr__(key_bundle._bundle, "iv_is_fresh", True)
                ciphertext = await cipher._Junction.abytes_encipher(
                    abatch(plaintext, size=config.BLOCKSIZE),
                    shmac=cipher._StreamHMAC(key_bundle)._for_encryption(),
                ).asend(None)
                aciphertexts.add(ciphertext)

            # the vulnerable first block of async ciphertexts is always
            # unique
            assert len(aciphertexts) == self.number_of_tests

            # the most vulnerable first INNER_HEADER-bytes of async
            # ciphertexts are also always unique
            ainner_headers = {
                aciphertext[config.INNER_HEADER_SLICE]
                for aciphertext in aciphertexts
            }
            assert len(ainner_headers) == self.number_of_tests

    async def test_sync_siv_only_resistance(self) -> None:
        for config, cipher, salt, _ in all_ciphers:
            iv = token_bytes(config.IV_BYTES)
            kw = dict(kdfs=cipher._kdfs, salt=salt, iv=iv)

            # aggregate the first ciphertext block of a collection of
            # ciphertexts instantiated with the same key, salt, aad & iv
            ciphertexts = set()
            for _ in range(self.number_of_tests):
                plaintext = cipher._padding.start_padding() + (
                    (config.BLOCKSIZE - config.INNER_HEADER_BYTES) * b"\x00"
                )
                key_bundle = cipher._KeyAADBundle(**kw).sync_mode()
                object.__setattr__(key_bundle._bundle, "iv_is_fresh", True)
                ciphertext = cipher._Junction.bytes_encipher(
                    batch(plaintext, size=config.BLOCKSIZE),
                    shmac=cipher._StreamHMAC(key_bundle)._for_encryption(),
                ).send(None)
                ciphertexts.add(ciphertext)

            # the vulnerable first block of ciphertexts is always unique
            assert len(ciphertexts) == self.number_of_tests

            # the most vulnerable first INNER_HEADER-bytes of ciphertexts
            # are also always unique
            inner_headers = {
                ciphertext[config.INNER_HEADER_SLICE]
                for ciphertext in ciphertexts
            }
            assert len(inner_headers) == self.number_of_tests

    async def test_async_single_component_resistance(self) -> None:
        for config, cipher, static_salt, static_aad in all_ciphers:
            static_iv = token_bytes(config.IV_BYTES)
            static_plaintext = config.BLOCKSIZE * b"\x00"
            kw = dict(
                kdfs=cipher._kdfs,
                salt=static_salt,
                iv=static_iv,
                aad=static_aad,
            )
            # aggregate first ciphertext blocks created with all static
            # randomizer components except one.
            for component, size in (
                ("salt", config.SALT_BYTES),
                ("aad", 16),
                ("iv", config.IV_BYTES),
            ):
                aciphertexts = set()

                for _ in range(self.number_of_tests):
                    key_bundle = await cipher._KeyAADBundle(
                        **{**kw, component: token_bytes(size)}
                    ).async_mode()
                    object.__setattr__(
                        key_bundle._bundle, "iv_is_fresh", True
                    )
                    aciphertext = await cipher._Junction.abytes_encipher(
                        abatch(static_plaintext, size=config.BLOCKSIZE),
                        shmac=cipher._StreamHMAC(
                            key_bundle
                        )._for_encryption(),
                    ).asend(None)
                    aciphertexts.add(aciphertext)

                # the vulnerable first block of ciphertexts is always unique
                assert len(aciphertexts) == self.number_of_tests

                # the most vulnerable first INNER_HEADER-bytes of ciphertexts
                # are also always unique
                inner_headers = {
                    aciphertext[config.INNER_HEADER_SLICE]
                    for aciphertext in aciphertexts
                }
                assert len(inner_headers) == self.number_of_tests

    async def test_sync_single_component_resistance(self) -> None:
        for config, cipher, static_salt, static_aad in all_ciphers:
            static_iv = token_bytes(config.IV_BYTES)
            static_plaintext = config.BLOCKSIZE * b"\x00"
            kw = dict(
                kdfs=cipher._kdfs,
                salt=static_salt,
                iv=static_iv,
                aad=static_aad,
            )
            # aggregate first ciphertext blocks created with all static
            # randomizer components except one.
            for component, size in (
                ("salt", config.SALT_BYTES),
                ("aad", 16),
                ("iv", config.IV_BYTES),
            ):
                ciphertexts = set()

                for _ in range(self.number_of_tests):
                    key_bundle = cipher._KeyAADBundle(
                        **{**kw, component: token_bytes(size)}
                    ).sync_mode()
                    object.__setattr__(
                        key_bundle._bundle, "iv_is_fresh", True
                    )
                    ciphertext = cipher._Junction.bytes_encipher(
                        batch(static_plaintext, size=config.BLOCKSIZE),
                        shmac=cipher._StreamHMAC(
                            key_bundle
                        )._for_encryption(),
                    ).send(None)
                    ciphertexts.add(ciphertext)

                # the vulnerable first block of ciphertexts is always unique
                assert len(ciphertexts) == self.number_of_tests

                # the most vulnerable first INNER_HEADER-bytes of ciphertexts
                # are also always unique
                inner_headers = {
                    ciphertext[config.INNER_HEADER_SLICE]
                    for ciphertext in ciphertexts
                }
                assert len(inner_headers) == self.number_of_tests


class TestCipherModes:
    async def test_key_bundle_must_be_set_to_either_async_or_sync(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "A cipher ran without a mode set on its key bundle."
        )
        for _, cipher, salt, aad in all_ciphers:
            key_bundle = cipher._KeyAADBundle(
                cipher._kdfs, salt=salt, aad=aad
            )
            assert not hasattr(key_bundle, "mode")
            with Ignore(t.KDFModeNotDeclared, if_else=violation(problem)):
                cipher._StreamHMAC(key_bundle)

    async def test_async_encryption_must_be_used_with_async_components(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "A cipher ran with mismatched component modes set."
        )
        for config, cipher, salt, aad in all_ciphers:
            data = b"test_data..."
            key_bundle = await cipher._KeyAADBundle(
                cipher._kdfs, salt=salt, aad=aad
            ).async_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_encryption()
            data = batch(
                cipher._padding.pad_plaintext(data), size=config.BLOCKSIZE
            )
            with Ignore(ValueError, if_else=violation(problem)):
                b"".join(cipher._Junction.bytes_encipher(data, shmac=shmac))

    async def test_sync_encryption_must_be_used_with_sync_components(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "A cipher ran with mismatched component modes set."
        )
        for config, cipher, salt, aad in all_ciphers:
            data = b"test_data..."
            key_bundle = cipher._KeyAADBundle(
                cipher._kdfs, salt=salt, aad=aad
            ).sync_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_encryption()
            data = abatch(
                cipher._padding.pad_plaintext(data), size=config.BLOCKSIZE
            )
            with Ignore(ValueError, if_else=violation(problem)):
                [
                    block
                    async for block in cipher._Junction.abytes_encipher(
                        data, shmac=shmac
                    )
                ]

    async def test_async_decryption_must_be_used_with_async_components(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "A cipher ran with mismatched component modes set."
        )
        for config, cipher, _, aad in all_ciphers:
            data = Ciphertext(
                cipher.bytes_encrypt(b"test_data..."), config=config
            )
            key_bundle = await cipher._KeyAADBundle(
                cipher._kdfs, salt=data.salt, aad=aad, iv=data.iv
            ).async_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_decryption()
            data = batch(data.ciphertext, size=config.BLOCKSIZE)
            with Ignore(ValueError, if_else=violation(problem)):
                b"".join(cipher._Junction.bytes_decipher(data, shmac=shmac))

    async def test_sync_decryption_must_be_used_with_sync_components(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "A cipher ran with mismatched component modes set."
        )
        for config, cipher, _, aad in all_ciphers:
            data = Ciphertext(
                cipher.bytes_encrypt(b"test_data..."), config=config
            )
            key_bundle = cipher._KeyAADBundle(
                cipher._kdfs, salt=data.salt, aad=aad, iv=data.iv
            ).sync_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_decryption()
            data = abatch(data.ciphertext, size=config.BLOCKSIZE)
            with Ignore(ValueError, if_else=violation(problem)):
                [
                    block
                    async for block in cipher._Junction.abytes_decipher(
                        data, shmac=shmac
                    )
                ]

    async def test_async_encryption_components_cant_be_used_for_decryption(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "A cipher ran with mismatched component modes set."
        )
        for config, cipher, salt, aad in all_ciphers:
            data = b"test_data..."
            key_bundle = await cipher._KeyAADBundle(
                cipher._kdfs, salt=salt, aad=aad
            ).async_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_encryption()
            data = abatch(
                cipher._padding.pad_plaintext(data), size=config.BLOCKSIZE
            )
            with Ignore(ValueError, if_else=violation(problem)):
                [
                    block
                    async for block in cipher._Junction.abytes_decipher(
                        data, shmac=shmac
                    )
                ]

    async def test_sync_encryption_components_cant_be_used_for_decryption(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "A cipher ran with mismatched component modes set."
        )
        for config, cipher, salt, aad in all_ciphers:
            data = b"test_data..."
            key_bundle = cipher._KeyAADBundle(
                cipher._kdfs, salt=salt, aad=aad
            ).sync_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_encryption()  # enc
            data = batch(
                cipher._padding.pad_plaintext(data), size=config.BLOCKSIZE
            )
            with Ignore(ValueError, if_else=violation(problem)):
                b"".join(cipher._Junction.bytes_decipher(data, shmac=shmac))

    async def test_async_decryption_components_cant_be_used_for_encryption(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "A cipher ran with mismatched component modes set."
        )
        for config, cipher, _, aad in all_ciphers:
            data = Ciphertext(
                cipher.bytes_encrypt(b"test_data..."), config=config
            )
            key_bundle = await cipher._KeyAADBundle(
                cipher._kdfs, salt=data.salt, aad=aad, iv=data.iv
            ).async_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_decryption()
            data = abatch(data.ciphertext, size=config.BLOCKSIZE)
            with Ignore(ValueError, if_else=violation(problem)):
                [
                    block
                    async for block in cipher._Junction.abytes_encipher(
                        data, shmac=shmac
                    )
                ]

    async def test_sync_decryption_components_cant_be_used_for_encryption(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "A cipher ran with mismatched component modes set."
        )
        for config, cipher, _, aad in all_ciphers:
            data = Ciphertext(
                cipher.bytes_encrypt(b"test_data..."), config=config
            )
            key_bundle = cipher._KeyAADBundle(
                cipher._kdfs, salt=data.salt, aad=aad, iv=data.iv
            ).sync_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_decryption()
            data = batch(data.ciphertext, size=config.BLOCKSIZE)
            with Ignore(ValueError, if_else=violation(problem)):
                b"".join(cipher._Junction.bytes_encipher(data, shmac=shmac))

    async def test_cant_set_encryption_mode_more_than_once(self) -> None:
        problem = (  # fmt: skip
            "Encryption mode was allowed to be set more than once."
        )
        for _, cipher, salt, aad in all_ciphers:
            key_bundle = await cipher._KeyAADBundle(
                cipher._kdfs, salt=salt, aad=aad
            ).async_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_encryption()
            with Ignore(PermissionError, if_else=violation(problem)):
                shmac._for_encryption()

            key_bundle = cipher._KeyAADBundle(
                cipher._kdfs, salt=salt, aad=aad
            ).sync_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_encryption()
            with Ignore(PermissionError, if_else=violation(problem)):
                shmac._for_encryption()

    async def test_cant_set_decryption_mode_without_setting_iv(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "Decryption mode was set without also setting IV."
        )
        for _, cipher, salt, aad in all_ciphers:
            key_bundle = await cipher._KeyAADBundle(
                cipher._kdfs, salt=salt, aad=aad
            ).async_mode()
            with Ignore(PermissionError, if_else=violation(problem)):
                cipher._StreamHMAC(key_bundle)._for_decryption()

            key_bundle = cipher._KeyAADBundle(
                cipher._kdfs, salt=salt, aad=aad
            ).sync_mode()
            with Ignore(PermissionError, if_else=violation(problem)):
                cipher._StreamHMAC(key_bundle)._for_decryption()

    async def test_cant_set_decryption_mode_more_than_once(self) -> None:
        problem = (  # fmt: skip
            "Decryption mode was allowed to be set more than once."
        )
        for config, cipher, salt, aad in all_ciphers:
            key_bundle = await cipher._KeyAADBundle(
                cipher._kdfs,
                salt=salt,
                aad=aad,
                iv=token_bytes(config.IV_BYTES),
            ).async_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_decryption()
            with Ignore(PermissionError, if_else=violation(problem)):
                shmac._for_decryption()

            key_bundle = cipher._KeyAADBundle(
                cipher._kdfs,
                salt=salt,
                aad=aad,
                iv=token_bytes(config.IV_BYTES),
            ).sync_mode()
            shmac = cipher._StreamHMAC(key_bundle)._for_decryption()
            with Ignore(PermissionError, if_else=violation(problem)):
                shmac._for_decryption()

    async def test_cant_manually_set_iv_and_encryption_mode(self) -> None:
        problem = (  # fmt: skip
            "A manually set IV was allowed in encryption mode."
        )
        for config, cipher, salt, aad in all_ciphers:
            key_bundle = await cipher._KeyAADBundle(
                cipher._kdfs,
                salt=salt,
                aad=aad,
                iv=token_bytes(config.IV_BYTES),
            ).async_mode()
            with Ignore(PermissionError, if_else=violation(problem)):
                cipher._StreamHMAC(key_bundle)._for_encryption()

            key_bundle = cipher._KeyAADBundle(
                cipher._kdfs,
                salt=salt,
                aad=aad,
                iv=token_bytes(config.IV_BYTES),
            ).sync_mode()
            with Ignore(PermissionError, if_else=violation(problem)):
                cipher._StreamHMAC(key_bundle)._for_encryption()


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Organizes the package's exceptions in declarative & expressive
classes & methods.
"""

__all__ = [
    "AuthenticationFailed",
    "CanonicalEncodingError",
    "CanonicalIssue",
    "CipherStreamIsClosed",
    "DatabaseNotConnected",
    "DatabaseIssue",
    "MissingDeclaredVariables",
    "Ignore",
    "ImproperPassphrase",
    "IncoherentConcurrencyState",
    "InvalidBlockID",
    "InvalidCiphertextSize",
    "InvalidDigest",
    "InvalidPassphrase",
    "InvalidSHMAC",
    "InvalidSignature",
    "Issue",
    "KDFModeNotDeclared",
    "KeyAADIssue",
    "Metadata",
    "PackageNotSigned",
    "PackageSignerIssue",
    "PasscryptIssue",
    "SHMACIssue",
    "SigningKeyNotSet",
    "TimestampExpired",
    "TypeUncheckableAtRuntime",
    "UndefinedRequiredAttributes",
    "ValidationIncomplete",
    "raise_exception",
]


import asyncio
from pathlib import Path
from cryptography.exceptions import InvalidSignature

from ._typing import Typing as t


def raise_exception(obj: Exception, /) -> None:
    """
    Simply provides a callable which raises `obj` turning the raise
    statement into an expression.
    """
    raise obj


class Metadata:
    """
    Creates efficient containers for the generic metadata of values.
    This is used so that potentially sensitive values can be analyzed
    while keeping the raw values from being passed around or displayed
    in error outputs.
    """

    __slots__ = ("size", "type")

    def __init__(self, value: t.Any, /) -> None:
        self.type = value.__class__
        self.size = len(value) if hasattr(value, "__len__") else None


class Ignore:
    """
    Allows specialized surpressing & handling of exceptions.
     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    with Ignore(TypeError):
        c = None + "Hello"
        # exception within context is surpressed if it's a `TypeError`

    Or, dynamically choose which exceptions to catch, and call custom
    cleanup code.

    def cleanup(relay: Ignore) -> bool:
        if issubclass(relay.error.__class__, PermissionError):
            return False
        else:
            db.save_database()
            return True

    errors = (IOError, PermissionError)
    with Ignore(*errors) as relay:
        relay.except_code = cleanup
        # Analogous to `with Ignore(*errors, if_except=cleanup):`
        # Runs `cleanup` if the specified type of error is raised. If
        # `cleanup` returns `True` the exception is surpressed.

        relay.else_code = cleanup
        # Analogous to `with Ignore(*errors, if_else=cleanup):`
        # Runs `cleanup` if no exception is raised in the context.

        relay.finally_code = cleanup
        # Analogous to `with Ignore(*errors, finally_run=cleanup):`
        # Always runs `cleanup` at the end of the context.

    async def acleanup(relay: Ignore) -> bool:
        ...

    async with Ignore(*errors, if_except=acleanup, ...):
        ...
    """

    __slots__ = (
        "ignored_exceptions",
        "except_code",
        "else_code",
        "finally_code",
        "bus",
        "error",
        "traceback",
    )

    class _PlaceholderHandler:
        """
        Stand-in handler when one isn't specified.
        """

        def __await__(self, /) -> t.Self:
            yield
            return self

        def __call__(self, /, *_: t.Any, **__: t.Any) -> t.Self:
            return self

        def __bool__(self, /) -> bool:
            return True

    def __init__(
        self,
        /,
        *exceptions: Exception,
        if_except: t.Optional[t.Callable[[t.Self], t.Any]] = None,
        if_else: t.Optional[t.Callable[[t.Self], t.Any]] = None,
        finally_run: t.Optional[t.Callable[[t.Self], t.Any]] = None,
    ) -> None:
        placeholder = self._PlaceholderHandler()
        self.ignored_exceptions = exceptions
        self.except_code = placeholder if if_except is None else if_except
        self.else_code = placeholder if if_else is None else if_else
        self.finally_code = (
            placeholder if finally_run is None else finally_run
        )
        self.bus = t.Namespace()
        self.error = None
        self.traceback = None

    def __repr__(self, /) -> str:
        return repr(getattr(self, "error", None))

    async def __aenter__(self, /) -> t.Self:
        """
        Open an async context.
        """
        await asyncio.sleep(0)
        return self

    def __enter__(self, /) -> t.Self:
        """
        Open a sync context.
        """
        return self

    async def __aexit__(
        self,
        /,
        exc_type: t.Optional[type] = None,
        exc_value: t.Optional[Exception] = None,
        traceback: t.Optional[t.TracebackType] = None,
    ) -> bool:
        """
        Allows the controlled handling of raised exceptions within the
        context. If an exception specified by the instance is raised,
        the `if_except` method must return a `bool`: returning `True`
        surpresses the propagation of the exception, returning `False`
        does not.
        """
        try:
            if exc_type is None:
                await self.else_code(self)
            else:
                self.error = exc_value
                self.traceback = traceback
                if issubclass(exc_type, self.ignored_exceptions):
                    return await self.except_code(self)
                else:
                    raise exc_value
        finally:
            await self.finally_code(self)

    def __exit__(
        self,
        /,
        exc_type: t.Optional[type] = None,
        exc_value: t.Optional[Exception] = None,
        traceback: t.Optional[t.TracebackType] = None,
    ) -> bool:
        """
        Allows the controlled handling of raised exceptions within the
        context. If an exception specified by the instance is raised,
        the `if_except` method must return a `bool`: returning `True`
        surpresses the propagation of the exception, returning `False`
        does not.
        """
        try:
            if exc_type is None:
                self.else_code(self)
            else:
                self.error = exc_value
                self.traceback = traceback
                if issubclass(exc_type, self.ignored_exceptions):
                    return self.except_code(self)
                else:
                    raise exc_value
        finally:
            self.finally_code(self)


class CanonicalEncodingError(ValueError):
    """
    An exception raised when a discrepancy is detected between the
    metadata declarations of canonically encoded data & the data itself.
    """


class TimestampExpired(TimeoutError):
    """
    An exception raised when an encountered timestamp is older than the
    current time minus a validation method's `ttl` parameter.
    """

    _DEFAULT_MESSAGE: str = "Timestamp expired by <TIME> UNITS."

    def __init__(self, units: str, expired_by: int, /, *a: t.Any) -> None:
        self.units = units
        self.expired_by = expired_by
        message = self._DEFAULT_MESSAGE.replace("UNITS", repr(units))
        super().__init__(message.replace("TIME", repr(expired_by)), *a)


class AuthenticationFailed(ValueError):
    """
    A base exception class for varying kinds of authentication failures.
    """


class InvalidBlockID(AuthenticationFailed):
    """
    An exception raised when a received `block_id` & `ciphertext_block`
    pair are invalidated by a `StreamHMAC` object.
    """


class InvalidSHMAC(AuthenticationFailed):
    """
    An exception raised when at the end of processing a stream of
    ciphertext, a `shmac` tag is invalidated by the `StreamHMAC` object.
    """


class InvalidDigest(AuthenticationFailed):
    """
    An exception raised when a purported checksum of data does not match
    the calculated checksum.
    """


class InvalidPassphrase(AuthenticationFailed):
    """
    An exception raised when the validation of a supplied passphrase
    fails.
    """


class ImproperPassphrase(ValueError):
    """
    If the passphrase supplied is either not bytes or not long enough
    this exception is raised.
    """


class UndefinedRequiredAttributes(AttributeError):
    """
    Captures the undefined attribute names within the raised exception.
    """

    __slots__ = ("undefined_attributes",)

    _MESSAGE_TEMPLATE: str = (
        "The following required attributes remained undefined after "
        "object initialization: UNDEFINED_ATTRIBUTES."
    )

    def __init__(self, /, *undefined_attributes: str) -> None:
        self.undefined_attributes = undefined_attributes
        # fmt: off
        super().__init__(
            self
            ._MESSAGE_TEMPLATE
            .replace("UNDEFINED_ATTRIBUTES", repr(undefined_attributes))
        )
        # fmt: on


class MissingDeclaredVariables(AttributeError):
    """
    Raised on mismatches between declared & supplied variables.
    """

    # fmt: off
    # TODO @rmlibre: switch templates throughout?
    # This style of template is safer with
    # untrusted inputs.
    _MESSAGE_TEMPLATE: str = (
        "The ",                                 # names
        " variable(s) were declared in the ",   # found_in
        " but were't detected in the ",         # missed
        "",
    )
    # fmt: on

    def __init__(self, /, *names: str, found_in: str, missed: str) -> None:
        msg = self._MESSAGE_TEMPLATE
        report = (repr(names), repr(found_in), repr(missed), ".")
        super().__init__(
            "".join(mold + part for mold, part in zip(msg, report))
        )


class InvalidCiphertextSize(ValueError):
    """
    Thrown with the invalid size stored in an attribute.
    """

    __slots__ = ("size",)

    _MESSAGE_TEMPLATE: str = (
        "The given ciphertext length of SIZE isn't a valid size."
    )

    def __init__(self, size: int, /, *a: t.Any) -> None:
        self.size = size
        message = self._MESSAGE_TEMPLATE.replace("SIZE", repr(size))
        super().__init__(message, *a)


class ValidationIncomplete(AttributeError):
    _MESSAGE_TEMPLATE: str = "Can't produce a result before finalization."

    def __init__(self, /, *a: t.Any) -> None:
        super().__init__(self._MESSAGE_TEMPLATE, *a)


class IncoherentConcurrencyState(RuntimeError):
    _MESSAGE_TEMPLATE: str = (
        "A concurrent environment has indicated out-of-order execution, "
        "leading to an incorrect internal state."
    )

    def __init__(self, /, *a: t.Any) -> None:
        super().__init__(self._MESSAGE_TEMPLATE, *a)


class CipherStreamIsClosed(InterruptedError):
    _MESSAGE_TEMPLATE: str = (
        "The stream has been closed. Cannot add more `data` to the "
        "buffer of an already closed stream."
    )

    def __init__(self, /, *a: t.Any) -> None:
        super().__init__(self._MESSAGE_TEMPLATE, *a)


class DatabaseNotConnected(AttributeError):
    _MESSAGE_TEMPLATE: str = (
        "Must first connect to the package signing session's secure "
        "database before it can be updated or queried."
    )

    def __init__(self, /, *a: t.Any) -> None:
        super().__init__(self._MESSAGE_TEMPLATE, *a)


class PackageNotSigned(AttributeError):
    _MESSAGE_TEMPLATE: str = (
        "This version of the package must be signed before querying its "
        "signature."
    )

    def __init__(self, /, *a: t.Any) -> None:
        super().__init__(self._MESSAGE_TEMPLATE, *a)


class SigningKeyNotSet(AttributeError):
    _MESSAGE_TEMPLATE: str = (
        "The `PackageSigner` instance's signing key hasn't been set."
    )

    def __init__(self, /, *a: t.Any) -> None:
        super().__init__(self._MESSAGE_TEMPLATE, *a)


class KDFModeNotDeclared(AttributeError):
    _MESSAGE_TEMPLATE: str = (
        "KeyAADBundle objects need to be set to either sync or async "
        "modes prior to querying their derived keys."
    )

    def __init__(self, /, *a: t.Any) -> None:
        super().__init__(self._MESSAGE_TEMPLATE, *a)


class TypeUncheckableAtRuntime(TypeError):
    """
    Some types, like those which use square brackets in their definition,
    cannot be checked with `isinstance`. This exception is raised when
    such types are defined on variables which need to be type-checked at
    runtime.
    """

    _MESSAGE_TEMPLATE: str = (
        "The NAME variable's type was declared using VALUE_TYPE, which "
        "isn't checkable at runtime. Perhaps use a `Protocol` decorated "
        "with `@typing.runtime_checkable` instead?"
    )

    def __init__(self, name: str, value_type: type, /, *a: t.Any) -> None:
        self.name = name
        self.value_type = value_type
        # fmt: off
        super().__init__(
            self
            ._MESSAGE_TEMPLATE
            .replace("NAME", repr(name))
            .replace("VALUE_TYPE", repr(value_type)),
            *a,
        )
        # fmt: on


class Issue:
    """
    Organizes messages for general / miscellaneous raised exceptions.
    """

    __slots__ = ()

    _INVALID_VALUE: str = "Invalid NAME!"
    _INVALID_LENGTH: str = "len(NAME) != LENGTH."
    _VALUE_MUST: str = "The NAME value must CONTEXT."
    _MUST_SET_VALUE: str = "Must set NAME for CONTEXT."
    _STREAM_IS_EMPTY: str = "An invalid emtpy stream was provided."
    _VALUE_ALREADY_SET: str = "The OBJECT is already set to CONTEXT."
    _VALUE_MUST_BE_TYPE: str = "The NAME value must be a TYPE object."
    _VALUE_MUST_BE_SUBTYPE: str = (
        "The NAME value must match or subclass TYPE."
    )
    _EXCEEDED_BLOCKSIZE: str = "Data block MUST NOT exceed BLOCKSIZE bytes."
    _CANT_REASSIGN_ATTRIBUTE: str = (
        "Can't re-assign the existing NAME attribute."
    )
    _CANT_DEASSIGN_ATTRIBUTE: str = (
        "Can't de-assign the existing NAME attribute."
    )
    _UNUSED_PARAMETERS: str = (
        "The PARAMETERS parameters are not used when CONTEXT."
    )
    _BROKEN_POOL_RESTARTED: str = (
        "The process pool was broken & has now been restarted. Try again."
    )

    @classmethod
    def invalid_value(cls, name: str, problem: str = "", /) -> ValueError:
        issue = cls._INVALID_VALUE.replace("NAME", name)
        if problem:
            issue = f"{issue} The {name} can't be {problem}."
        return ValueError(issue)

    @classmethod
    def invalid_length(cls, name: str, length: int, /) -> ValueError:
        issue = cls._INVALID_LENGTH.replace("NAME", name)
        return ValueError(issue.replace("LENGTH", repr(length)))

    @classmethod
    def value_must(cls, name: str, context: t.Any, /) -> ValueError:
        issue = cls._VALUE_MUST.replace("NAME", repr(name))
        return ValueError(issue.replace("CONTEXT", str(context)))

    @classmethod
    def must_set_value(cls, name: str, context: str, /) -> ValueError:
        issue = cls._MUST_SET_VALUE.replace("NAME", repr(name))
        return ValueError(issue.replace("CONTEXT", str(context)))

    @classmethod
    def stream_is_empty(cls, /) -> ValueError:
        return ValueError(cls._STREAM_IS_EMPTY)

    @classmethod
    def value_already_set(
        cls, obj: str, context: str, /
    ) -> PermissionError:
        issue = cls._VALUE_ALREADY_SET.replace("OBJECT", str(obj))
        return PermissionError(issue.replace("CONTEXT", str(context)))

    @classmethod
    def value_must_be_type(cls, name: str, clss: t.Any, /) -> TypeError:
        issue = cls._VALUE_MUST_BE_TYPE.replace("NAME", repr(name))
        return TypeError(issue.replace("TYPE", repr(clss)))

    @classmethod
    def value_must_be_subtype(cls, name: str, clss: t.Any, /) -> TypeError:
        issue = cls._VALUE_MUST_BE_SUBTYPE.replace("NAME", repr(name))
        return TypeError(issue.replace("TYPE", repr(clss)))

    @classmethod
    def exceeded_blocksize(cls, blocksize: int, /) -> OverflowError:
        issue = cls._EXCEEDED_BLOCKSIZE
        return OverflowError(issue.replace("BLOCKSIZE", repr(blocksize)))

    @classmethod
    def cant_reassign_attribute(cls, name: str, /) -> PermissionError:
        issue = cls._CANT_REASSIGN_ATTRIBUTE
        return PermissionError(issue.replace("NAME", repr(name)))

    @classmethod
    def cant_deassign_attribute(cls, name: str, /) -> PermissionError:
        issue = cls._CANT_DEASSIGN_ATTRIBUTE
        return PermissionError(issue.replace("NAME", repr(name)))

    @classmethod
    def unused_parameters(
        cls, params: t.Any, context: str, /
    ) -> ValueError:
        issue = cls._UNUSED_PARAMETERS.replace("PARAMETERS", repr(params))
        return ValueError(issue.replace("CONTEXT", str(context)))

    @classmethod
    def broken_pool_restarted(cls, /) -> RuntimeError:
        return RuntimeError(cls._BROKEN_POOL_RESTARTED)  # pragme: no cover


class CanonicalIssue:
    """
    Organizes messages for raised exceptions related to canonicalizing
    collections of bytes data.
    """

    __slots__ = ()

    _INFLATED_SIZE_DECLARATION: str = (
        "More items to unpack than data permits!"
    )
    _ITEM_LENGTH_MISMATCH: str = (
        "The measured length of the canonically encoded item does not "
        "match its declared length."
    )
    _INVALID_PADDING: str = "Invalid canonical encoding padding detected!"
    _DATA_LENGTH_BLOCKSIZE_MISMATCH: str = (
        "Multiple of data length != declared blocksize!"
    )
    _MISSING_METADATA_ITEMS: str = (
        "The encoding lead to an unpacked data result without its "
        "required blocksize / pad declarations."
    )

    @classmethod
    def item_length_mismatch(cls, /) -> CanonicalEncodingError:
        return CanonicalEncodingError(cls._ITEM_LENGTH_MISMATCH)

    @classmethod
    def inflated_size_declaration(cls, /) -> CanonicalEncodingError:
        return CanonicalEncodingError(cls._INFLATED_SIZE_DECLARATION)

    @classmethod
    def invalid_padding(cls, /) -> CanonicalEncodingError:
        return CanonicalEncodingError(cls._INVALID_PADDING)

    @classmethod
    def data_length_blocksize_mismatch(cls, /) -> CanonicalEncodingError:
        return CanonicalEncodingError(cls._DATA_LENGTH_BLOCKSIZE_MISMATCH)

    @classmethod
    def missing_metadata_items(cls, /) -> CanonicalEncodingError:
        return CanonicalEncodingError(cls._MISSING_METADATA_ITEMS)


class KeyAADIssue:
    """
    Organizes messages for raised exceptions related to the invalid
    usage or structure of keys & associated session information.
    """

    __slots__ = ()

    _INVALID_KEY_SIZE: str = (
        "`key` was KEY_SIZE bytes, but must be at least MIN_SIZE bytes."
    )
    _ALREADY_REGISTERED: str = "The shmac has already been registered."
    _MODE_ISNT_CORRECT: str = (
        "The KDF mode must be set to MODE to use MODE key derivation."
    )

    KDFModeNotDeclared: type = KDFModeNotDeclared

    @classmethod
    def invalid_key_size(cls, size: int, min_size: int, /) -> ValueError:
        issue = cls._INVALID_KEY_SIZE.replace("MIN_SIZE", repr(min_size))
        return ValueError(issue.replace("KEY_SIZE", repr(size)))

    @classmethod
    def shmac_already_registered(cls, /) -> PermissionError:
        return PermissionError(cls._ALREADY_REGISTERED)

    @classmethod
    def mode_isnt_correct(cls, mode: str, /) -> ValueError:
        issue = cls._MODE_ISNT_CORRECT.replace("MODE", repr(mode))
        return ValueError(issue)


class SHMACIssue:
    """
    Organizes messages for raised exceptions related to `StreamHMAC`.
    """

    __slots__ = ()

    _NO_CIPHER_MODE_DECLARED: str = "No cipher mode has been declared."
    _ALREADY_FINALIZED: str = "The validator has already been finalized."
    _INVALID_IV_USAGE: str = (
        "The `iv` must be manually passed into the validator during "
        "*decryption*, & only during decryption."
    )
    _BLOCK_ID_IS_TOO_SMALL: str = (
        "A block id of SIZE bytes is too small. It must be at least "
        "MIN bytes to securely authenticate a block."
    )
    _BLOCK_ID_IS_TOO_BIG: str = (
        "A block id of SIZE bytes is too big. It can be at most MAX "
        "bytes."
    )
    _INVALID_SHMAC: str = (
        "Invalid StreamHMAC hash for the given ciphertext."
    )
    _INVALID_BLOCK_ID: str = (
        "Invalid next block ID hash of the supplied ciphertext block."
    )

    @classmethod
    def no_cipher_mode_declared(cls, /) -> PermissionError:
        return PermissionError(cls._NO_CIPHER_MODE_DECLARED)

    @classmethod
    def already_finalized(cls, /) -> PermissionError:
        return PermissionError(cls._ALREADY_FINALIZED)

    @classmethod
    def invalid_iv_usage(cls, /) -> PermissionError:
        return PermissionError(cls._INVALID_IV_USAGE)

    @classmethod
    def block_id_is_too_small(
        cls, size: int, min_size: int, /
    ) -> PermissionError:
        issue = cls._BLOCK_ID_IS_TOO_SMALL.replace("SIZE", repr(size))
        issue = issue.replace("MIN", repr(min_size))
        return PermissionError(issue)

    @classmethod
    def block_id_is_too_big(
        cls, size: int, max_size: int, /
    ) -> PermissionError:
        issue = cls._BLOCK_ID_IS_TOO_BIG.replace("SIZE", repr(size))
        issue = issue.replace("MAX", repr(max_size))
        return PermissionError(issue)

    @classmethod
    def invalid_shmac(cls, /) -> InvalidSHMAC:
        return InvalidSHMAC(cls._INVALID_SHMAC)

    @classmethod
    def invalid_block_id(cls, /) -> InvalidBlockID:
        return InvalidBlockID(cls._INVALID_BLOCK_ID)


class PasscryptIssue:
    """
    Organizes messages for raised exceptions related to `Passcrypt`.
    """

    __slots__ = ()

    _IMPROPER_PASSPHRASE: str = (
        "The given passphrase is too short, at least NEED more characters "
        "need to be added. \nTry using aiootp.mnemonic, it can help in "
        "generating a strong passphrase.\n"
        " _____________________________________\n"
        "|                                     |\n"
        "|            Usage Example:           |\n"
        "|_____________________________________|\n\n"
        "my_new_passphrase = b'-'.join(aiootp.mnemonic())\n\n"
        "print(my_new_passphrase)\n"
        "b'review-letter-blast-giant-connect-ring-balcony-frown'"
    )
    _IMPROPER_SALT: str = (
        "len(salt) must be >= MIN_SALT_SIZE and <= MAX_SALT_SIZE"
    )
    _INVALID_MB: str = "mb:MB must be int >= MIN_MB and <= MAX_MB"
    _INVALID_CPU: str = "cpu:CPU must be int >= MIN_CPU and <= MAX_CPU"
    _INVALID_CORES: str = (
        "cores:CORES must be int >= MIN_CORES and <= MAX_CORES"
    )
    _INVALID_TAG_SIZE: str = "tag_size:SIZE must be int >= MIN_TAG_SIZE"
    _INVALID_SALT_SIZE: str = (
        "salt_size:SIZE must be int >= MIN_SALT_SIZE and <= MAX_SALT_SIZE"
    )
    _UNTRUSTED_RESOURCE_CONSUMPTION: str = (
        "The PARAMETER parameter was blocked from being processed because "
        "it fell outside of the allowed range of resource consumption set "
        "by the verification method. To continue verifying the provided "
        "hash, explicit permission must be given in the form of a "
        "`builtins.range` object which includes the value VALUE.\n\n"
        " _____________________________________\n"
        "|                                     |\n"
        "|            Usage Example:           |\n"
        "|_____________________________________|\n\n"
        "from aiootp import Passcrypt\n\n"
        "allowed_resource_consumption = dict(\n"
        "    mb_allowed=range(16, 256),  # Less than 256 MiB allowed\n"
        "    cpu_allowed=range(2, 8),    # Less than 8 complexity allowed\n"
        "    cores_allowed=range(1, 5),  # Less than 5 processes allowed\n"
        ")\n"
        "try:\n"
        "    Passcrypt.verify(hashed_pw, pw, **allowed_resource_consumption)\n"
        "except ResourceWarning as danger:\n"
        "    admin.log(danger)\n"
        "    hard_limits_exceeded = (\n"
        "        danger.requested_resources.mb > 512\n"
        "        or danger.requested_resources.cpu > 11\n"
        "        or danger.requested_resources.cores > 8\n"
        "    )\n"
        "    below_security_guidelines = (\n"
        "        danger.requested_resources.mb < 16\n"
        "        or danger.requested_resources.cpu < 2\n"
        "    )\n"
        "    if hard_limits_exceeded:\n"
        "        raise danger\n"
        "    elif below_security_guidelines:\n"
        "        raise PermissionError('Minimum hash difficulty unmet.')\n"
        "    Passcrypt.verify(hashed_pw, pw)\n"
    )
    _VERIFICATION_FAILED: str = (
        "Passphrase verification failed! The hash of the passphrase & "
        "the passcrypt hash did not match!"
    )

    @classmethod
    def improper_passphrase(cls, metadata: Metadata, /) -> ValueError:
        from .keygens.passcrypt.config import passcrypt_spec as c

        if metadata.type is not bytes:
            return Issue.value_must_be_type("`passphrase`", bytes)
        issue = cls._IMPROPER_PASSPHRASE
        missing = c.MIN_PASSPHRASE_BYTES - metadata.size
        error = ImproperPassphrase(issue.replace("NEED", str(missing)))
        error.missing = missing
        return error

    @classmethod
    def improper_salt(cls, metadata: Metadata, /) -> ValueError:
        from .keygens.passcrypt.config import passcrypt_spec as c

        if metadata.type is not bytes:
            return Issue.value_must_be_type("`salt`", bytes)
        # fmt: off
        return ValueError(
            cls
            ._IMPROPER_SALT
            .replace("MIN_SALT_SIZE", repr(c.MIN_SALT_SIZE))
            .replace("MAX_SALT_SIZE", repr(c.MAX_SALT_SIZE))
        )
        # fmt: on

    @classmethod
    def improper_aad(cls, /) -> TypeError:
        return Issue.value_must_be_type("`aad`", bytes)

    @classmethod
    def invalid_mb(cls, mb: int, /) -> ValueError:
        from .keygens.passcrypt.config import passcrypt_spec as c

        if mb.__class__ is not int:
            return Issue.value_must_be_type("`mb`", int)
        # fmt: off
        return ValueError(
            cls
            ._INVALID_MB
            .replace("MIN_MB", repr(c.MIN_MB))
            .replace("MAX_MB", repr(c.MAX_MB))
            .replace("MB", repr(mb))
        )
        # fmt: on

    @classmethod
    def invalid_cpu(cls, cpu: int, /) -> ValueError:
        from .keygens.passcrypt.config import passcrypt_spec as c

        if cpu.__class__ is not int:
            return Issue.value_must_be_type("`cpu`", int)
        # fmt: off
        return ValueError(
            cls
            ._INVALID_CPU
            .replace("MIN_CPU", repr(c.MIN_CPU))
            .replace("MAX_CPU", repr(c.MAX_CPU))
            .replace("CPU", repr(cpu))
        )
        # fmt: on

    @classmethod
    def invalid_cores(cls, cores: int, /) -> ValueError:
        from .keygens.passcrypt.config import passcrypt_spec as c

        if cores.__class__ is not int:
            return Issue.value_must_be_type("`cores`", int)
        # fmt: off
        return ValueError(
            cls
            ._INVALID_CORES
            .replace("MIN_CORES", repr(c.MIN_CORES))
            .replace("MAX_CORES", repr(c.MAX_CORES))
            .replace("CORES", repr(cores))
        )
        # fmt: on

    @classmethod
    def invalid_tag_size(cls, tag_size: int, /) -> ValueError:
        from .keygens.passcrypt.config import passcrypt_spec as c

        if tag_size.__class__ is not int:
            return Issue.value_must_be_type("`tag_size`", int)
        # fmt: off
        return ValueError(
            cls
            ._INVALID_TAG_SIZE
            .replace("MIN_TAG_SIZE", repr(c.MIN_TAG_SIZE))
            .replace("SIZE", repr(tag_size))
        )
        # fmt: on

    @classmethod
    def invalid_salt_size(cls, salt_size: int, /) -> ValueError:
        from .keygens.passcrypt.config import passcrypt_spec as c

        if salt_size.__class__ is not int:
            return Issue.value_must_be_type("`salt_size`", int)
        # fmt: off
        return ValueError(
            cls
            ._INVALID_SALT_SIZE
            .replace("MIN_SALT_SIZE", repr(c.MIN_SALT_SIZE))
            .replace("MAX_SALT_SIZE", repr(c.MAX_SALT_SIZE))
            .replace("SIZE", repr(salt_size))
        )
        # fmt: on

    @classmethod
    def untrusted_resource_consumption(
        cls, parameter: str, header: t.Mapping[str, int], /
    ) -> ResourceWarning:
        value = header[parameter]
        issue = cls._UNTRUSTED_RESOURCE_CONSUMPTION
        issue = issue.replace("PARAMETER", repr(parameter))
        danger = ResourceWarning(issue.replace("VALUE", repr(value)))
        danger.requested_resources = header
        danger.parameter = parameter
        danger.value = value
        return danger

    @classmethod
    def verification_failed(cls, /) -> InvalidPassphrase:
        return InvalidPassphrase(cls._VERIFICATION_FAILED)


class DatabaseIssue:
    """
    Organizes messages for raised exceptions related to the classes
    `AsyncDatabase` & `Database`.
    """

    __slots__ = ()

    _FILE_NOT_FOUND: str = "The NAME filename was not located."
    _NO_EXISTING_METATAG: str = "No metatag database named TAG."
    _MISSING_PROFILE: str = "Profile doesn't exist or is corrupt."
    _TAG_FILE_DOESNT_EXIST: str = "TAG tag data isn't in the database."

    @classmethod
    def file_not_found(cls, filename: str, /) -> LookupError:
        issue = cls._FILE_NOT_FOUND
        return LookupError(issue.replace("NAME", repr(filename)))

    @classmethod
    def no_existing_metatag(cls, tag: str, /) -> LookupError:
        issue = cls._NO_EXISTING_METATAG
        return LookupError(issue.replace("TAG", repr(tag)))

    @classmethod
    def tag_file_doesnt_exist(cls, tag: str, /) -> LookupError:
        issue = cls._TAG_FILE_DOESNT_EXIST
        return LookupError(issue.replace("TAG", repr(tag)))


class PackageSignerIssue:
    """
    Organizes messages for raised exceptions related to  `PackageSigner`.
    """

    __slots__ = ()

    _INVALID_FILE_DIGEST: str = (
        "The summary & the hash digest of the given file don't match: "
        "FILENAME."
    )
    _UNSYNCED_PACKAGE_SIGNATURE: str = (
        "The calculated package signature is out of sync with the "
        "current checksum of the package summary."
    )

    DatabaseNotConnected: type = DatabaseNotConnected
    InvalidSignature: type = InvalidSignature
    PackageNotSigned: type = PackageNotSigned
    SigningKeyNotSet: type = SigningKeyNotSet

    @classmethod
    def invalid_file_digest(
        cls, filename: t.Union[str, Path], /
    ) -> InvalidDigest:
        issue = cls._INVALID_FILE_DIGEST
        return InvalidDigest(issue.replace("FILENAME", repr(filename)))

    @classmethod
    def unsynced_package_signature(cls, /) -> ValueError:
        return ValueError(cls._UNSYNCED_PACKAGE_SIGNATURE)


module_api = dict(
    AuthenticationFailed=t.add_type(AuthenticationFailed),
    CanonicalEncodingError=t.add_type(CanonicalEncodingError),
    CanonicalIssue=t.add_type(CanonicalIssue),
    CipherStreamIsClosed=t.add_type(CipherStreamIsClosed),
    DatabaseNotConnected=t.add_type(DatabaseNotConnected),
    DatabaseIssue=t.add_type(DatabaseIssue),
    MissingDeclaredVariables=t.add_type(MissingDeclaredVariables),
    Ignore=t.add_type(Ignore),
    ImproperPassphrase=t.add_type(ImproperPassphrase),
    IncoherentConcurrencyState=t.add_type(IncoherentConcurrencyState),
    InvalidBlockID=t.add_type(InvalidBlockID),
    InvalidCiphertextSize=t.add_type(InvalidCiphertextSize),
    InvalidDigest=t.add_type(InvalidDigest),
    InvalidPassphrase=t.add_type(InvalidPassphrase),
    InvalidSHMAC=t.add_type(InvalidSHMAC),
    InvalidSignature=t.add_type(InvalidSignature),
    Issue=t.add_type(Issue),
    KDFModeNotDeclared=t.add_type(KDFModeNotDeclared),
    KeyAADIssue=t.add_type(KeyAADIssue),
    Metadata=t.add_type(Metadata),
    PackageNotSigned=t.add_type(PackageNotSigned),
    PackageSignerIssue=t.add_type(PackageSignerIssue),
    PasscryptIssue=t.add_type(PasscryptIssue),
    SHMACIssue=t.add_type(SHMACIssue),
    SigningKeyNotSet=t.add_type(SigningKeyNotSet),
    TimestampExpired=t.add_type(TimestampExpired),
    TypeUncheckableAtRuntime=t.add_type(TypeUncheckableAtRuntime),
    UndefinedRequiredAttributes=t.add_type(UndefinedRequiredAttributes),
    ValidationIncomplete=t.add_type(ValidationIncomplete),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    raise_exception=raise_exception,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


import platform

from conftest import *


class BasicTestSuite:
    _type: type
    _id_name: str

    system = platform.system()

    async def aget_ids(self) -> t.Dict[str, int]:
        from aiootp.asynchs import asleep
        from aiootp.asynchs.threads import get_thread_id
        from aiootp.asynchs.processes import get_process_id

        await asleep()
        return dict(process_id=get_process_id(), thread_id=get_thread_id())

    def get_ids(self) -> t.Dict[str, int]:
        from aiootp.asynchs.threads import get_thread_id
        from aiootp.asynchs.processes import get_process_id

        return dict(process_id=get_process_id(), thread_id=get_thread_id())

    async def test_anew(self) -> None:
        name = self._id_name

        is_non_linux_multiprocessing_issue = lambda _: (
            (self._type is Processes) and (self.system != "Linux")
        )
        with Ignore(
            IndexError, if_except=is_non_linux_multiprocessing_issue
        ):
            result = (await self._type.anew(self.aget_ids))[name]
            assert result > 0
            assert result.__class__ is int
            assert result != (await self.aget_ids())[name]
            assert result != self.get_ids()[name]

            result = (await self._type.anew(self.get_ids))[name]
            assert result > 0
            assert result.__class__ is int
            assert result != self.get_ids()[name]

    def test_new(self) -> None:
        name = self._id_name

        is_non_linux_multiprocessing_issue = lambda _: (
            (self._type is Processes) and (self.system != "Linux")
        )
        with Ignore(
            IndexError, if_except=is_non_linux_multiprocessing_issue
        ):
            result = self._type.new(self.get_ids)[name]
            assert result > 0
            assert result.__class__ is int
            assert result != self.get_ids()[name]

    async def test_asubmit(self) -> None:
        name = self._id_name
        self._type.reset_pool()

        fut = await self._type.asubmit(self.aget_ids)
        result = (await fut.aresult())[name]
        assert result == fut.result()[name]
        assert result > 0
        assert result.__class__ is int
        assert result != (await self.aget_ids())[name]
        assert result != self.get_ids()[name]

        fut = await self._type.asubmit(self.get_ids)
        result = (await fut.aresult())[name]
        assert result == fut.result()[name]
        assert result > 0
        assert result.__class__ is int
        assert result != (await self.aget_ids())[name]
        assert result != self.get_ids()[name]

    def test_submit(self) -> None:
        name = self._id_name
        self._type.reset_pool()

        fut = self._type.submit(self.get_ids)
        result = fut.result()[name]
        assert result > 0
        assert result.__class__ is int
        assert result != self.get_ids()[name]

    async def test_probe_delay_must_be_positive(self) -> None:
        problem = (  # fmt: skip
            "A non-positive probe_delay was allowed."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            await self._type.anew(acsprng, probe_delay=-1)

        with Ignore(ValueError, if_else=violation(problem)):
            self._type.new(csprng, probe_delay=-1)


if platform.system() != "Windows":

    class TestProcesses(BasicTestSuite):
        _type: type = Processes
        _id_name: str = "process_id"


class TestThreads(BasicTestSuite):
    _type: type = Threads
    _id_name: str = "thread_id"


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *


class TestSequenceIDConfig:
    kw = dict(config_id=16, permutation_type=t.FastAffineXORChain)

    async def test_permutation_cid_defaults_to_instance_cid(self) -> None:
        config = t.SequenceIDConfig(size=16, **self.kw)
        assert 16 == config.PERMUTATION_CONFIG_ID

    async def test_size_must_be_within_bounded_limits(self) -> None:
        problem = (  # fmt: skip
            "A size out of bounds was allowed."
        )
        for bad_size in (-1, 0, 4097):
            with Ignore(ValueError, if_else=violation(problem)):
                t.SequenceIDConfig(size=bad_size, **self.kw)


class TestSequenceID:
    async def test_inversion_correctness(self) -> None:
        for n in range(1, 33):
            for index in range(0, 256, 16):
                key = token_bytes(SequenceID.key_size(n))
                sid = SequenceID(key=key, config_id=n)
                aresult = await sid.anew(index)
                result = sid.new(index)
                assert index == await sid.aread(aresult)
                assert index == await sid.aread(result)
                assert index == sid.read(aresult)
                assert index == sid.read(result)

    async def test_range_uniqueness(self) -> None:
        n = 1
        key = token_bytes(SequenceID.key_size(n))
        sid = SequenceID(key=key, config_id=n)

        # the sequence ids produced are unique for every given sequential
        # integer up to the max of the byte domain.
        history = set()
        for index in range(256**n):
            result = sid.new(index)
            assert index == sid.read(result)
            assert result not in history, f"{index=}, {n=}"
            history.add(result)

        # assert values wrap after the domain has been exhausted
        assert sid.new(index + 1) in history
        assert sid.new(index + 1) == await sid.anew(index + 1)

    async def test_sizes(self) -> None:
        for n in range(1, 33):
            # the salt must be at least the length of output sizes
            problem = (  # fmt: skip
                "An invalid key size for a specified config_id was allowed."
            )
            with Ignore(KeyError, ValueError, if_else=violation(problem)):
                sid = SequenceID(
                    key=token_bytes(SequenceID.key_size(n - 1)), config_id=n
                )

            sid = SequenceID(
                key=token_bytes(SequenceID.key_size(n)), config_id=n
            )
            # the size of produced sequential ids is the same as the defined
            # size
            result = sid.new(n)
            assert len(result) == n
            assert type(result) is bytes
            assert result == await sid.anew(n)


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Type-hinting utilities for the package.
"""

__all__ = ["Typing"]


from .interface import *
from ._commons import *
from ._generics import *
from ._asynchs import *
from ._paths import *
from ._randoms import *
from ._permutations import *
from ._ciphers import *
from ._keygens import *
from ._databases import *
from .tuples import *


modules = dict(
    _asynchs=_asynchs,
    _ciphers=_ciphers,
    _commons=_commons,
    _databases=_databases,
    _generics=_generics,
    _keygens=_keygens,
    _paths=_paths,
    _permutations=_permutations,
    _randoms=_randoms,
    interface=interface,
    tuples=tuples,
)


module_api = dict(
    Typing=Typing,
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Implementations of simple keyed permutations.
"""

__all__ = ["FastAffineXORChain"]


from .affine_permutation_config import *
from .affine_permutation import *
from .affine_xor_chain_config import *
from .affine_xor_chain import *
from .fast_affine_xor_chain import *


modules = dict(
    affine_permutation_config=affine_permutation_config,
    affine_permutation=affine_permutation,
    affine_xor_chain=affine_xor_chain,
    affine_xor_chain_config=affine_xor_chain_config,
    fast_affine_xor_chain=fast_affine_xor_chain,
)


module_api = dict(
    FastAffineXORChain=FastAffineXORChain,
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *

from aiootp._typing import tuples
from aiootp.commons.instances import *
from aiootp.commons.slots import *
from aiootp.commons.typed_slots import *
from aiootp.commons.namespaces import *
from aiootp.commons.configs import *


class SlotsAttributes:
    __slots__ = ()

    _items: t.Dict[str, t.Any] = dict(
        _private=True, one=1, mapped="value", unmapped=tuple("attr")
    )
    _MAPPED_ATTRIBUTES: t.Tuple[str] = tuple(
        name for name in _items if name != "unmapped"
    )
    _UNMAPPED_ATTRIBUTES: t.Tuple[str] = (
        *Slots._UNMAPPED_ATTRIBUTES,
        "mapped",
        "unmapped",
    )  # Being mapped overrides being unmapped


class NamespaceAttributes(SlotsAttributes):
    __slots__ = ()

    _UNMAPPED_ATTRIBUTES: t.Tuple[str] = (
        *Namespace._UNMAPPED_ATTRIBUTES,
        "mapped",
        "unmapped",
    )  # Being mapped overrides being unmapped


class BaseVariableHoldingClassTests:
    _items = dict(SlotsAttributes._items)
    _type: type
    _open: bool
    _frozen: bool

    async def test_init_accepts_mapping_and_keywords(self) -> None:
        obj_from_mapping = self._type(self._items)
        obj_from_keywords = self._type(**self._items)
        for name, value in self._items.items():
            assert value == getattr(obj_from_mapping, name)
            assert value == getattr(obj_from_keywords, name)

    async def test_attribute_access(self) -> None:
        obj = self._type()
        obj.mapped = self._items["mapped"]
        assert obj.mapped == self._items["mapped"]

        if_else = lambda _: not self._frozen or raise_exception(_.error)
        if_except = lambda _: self._frozen or raise_exception(_.error)
        with Ignore(PermissionError, if_else=if_else, if_except=if_except):
            del obj.mapped

        assert self._frozen or not hasattr(obj, "mapped")

    async def test_item_access(self) -> None:
        obj = self._type()

        try:
            obj[123] = True
        except (TypeError, AttributeError) as feature_broken:
            if hasattr(obj, "__dict__"):
                raise feature_broken
            else:
                return
        assert 123 in obj
        assert obj[123] is True

        if_except = lambda _: self._frozen
        if_else = lambda _: not self._frozen or raise_exception(_.error)
        with Ignore(PermissionError, if_except=if_except, if_else=if_else):
            del obj[123]

        assert self._frozen or 123 not in obj


class BaseReprControlledTests(BaseVariableHoldingClassTests):
    async def test_instance_repr_shows_values(self) -> None:
        DebugControl.disable_debugging()
        obj = self._type(self._items)
        string = repr(obj)
        for name, value in self._items.items():
            if (
                name[0] == "_"
                or hasattr(obj, "_is_mapped_attribute")
                and not obj._is_mapped_attribute(name)
            ):
                assert name not in string
            else:
                assert name in string, name
                assert self._open ^ (str(value) not in string), value
                assert self._open ^ (str(value.__class__) in string), value


class BaseMaskableReprTests(BaseVariableHoldingClassTests):
    async def test_mask_hides_or_shows_repr(self) -> None:
        DebugControl.disable_debugging()
        obj = self._type(self._items)
        for name in self._items:
            if str(name) in getattr(obj, "_UNMAPPED_ATTRIBUTES", ()) or str(
                name
            ).startswith("_"):
                continue
            assert str(name) in obj.__repr__(mask=False), name
            assert str(getattr(obj, name)) in obj.__repr__(mask=False), name
            assert str(name) in obj.__repr__(mask=True), name
            assert str(getattr(obj, name)) not in obj.__repr__(
                mask=True
            ), name
            DebugControl.enable_debugging(silence_warnings=True)
            assert str(name) in repr(obj), name
            assert str(getattr(obj, name)) in repr(obj), name
            DebugControl.disable_debugging()


class BaseFrozenTests(BaseVariableHoldingClassTests):
    async def test_instance_values_cant_be_changed_once_set(self) -> None:
        problem = (  # fmt: skip
            "An instance was allowed to be mutated."
        )
        obj = self._type()
        for name, value in self._items.items():
            setattr(obj, name, value)
            with Ignore(PermissionError, if_else=violation(problem)):
                setattr(obj, name, value)
            with Ignore(PermissionError, if_else=violation(problem)):
                delattr(obj, name)


class BaseDictLikeTests(BaseVariableHoldingClassTests):
    def frozen_violation_catcher(self, name: str) -> Ignore:
        if_except = lambda _: self._frozen
        if_else = lambda _: (
            (not self._frozen) or raise_exception(AssertionError(name))
        )
        return Ignore(PermissionError, if_except=if_except, if_else=if_else)

    async def test_inclusion_exclusion_logic(self) -> None:
        items = {**self._items, 123: "number"}
        obj = self._type()
        unmapped = set()
        MAPPED = getattr(obj, "_MAPPED_ATTRIBUTES", ())
        UNMAPPED = getattr(obj, "_UNMAPPED_ATTRIBUTES", ())
        for i, (name, value) in enumerate(items.items(), start=1):
            if name in MAPPED or name not in UNMAPPED:
                pass
            else:
                unmapped.add(name)
            obj[name] = value
            assert name in obj
            assert value == obj[name]
            assert i == len(obj) + len(unmapped)
            with self.frozen_violation_catcher(name):
                del obj[name]
                assert name not in obj
            with self.frozen_violation_catcher(name):
                obj[name] = value

    async def test_mapping_methods(self) -> None:
        items = {**self._items, 123: "number"}
        obj = self._type()
        for name, value in items.items():
            obj[name] = value

        assert set(items.keys()).issuperset(obj.keys())
        assert set(items.values()).issuperset(obj.values())
        assert set(items.items()).issuperset(obj.items())

        items.update(dict(new_value=True))
        obj.update(dict(new_value=True))
        assert obj["new_value"] is True
        assert obj["new_value"] == items["new_value"]

        with self.frozen_violation_catcher("new_value"):
            obj.update(dict(new_value=True).items())
            assert obj["new_value"] is True
        with self.frozen_violation_catcher("new_value"):
            obj.update(**dict(new_value=False))
            assert obj["new_value"] is False
        with self.frozen_violation_catcher("new_value"):
            obj.update(dict(new_value=False), **dict(new_value=True))
            assert obj["new_value"] is True

    async def test_strange_slots_dict_interplay_exists_but_is_avoided(
        self,
    ) -> None:
        if not hasattr(self._type, "__slots__"):
            return

        try:

            class MisusedSubclass(self._type):
                __slots__ = ("attr_0", "attr_1", "__dict__")
                slots_types = dict(attr_0=int, attr_1=int)
        except TypeError:

            class MisusedSubclass(self._type):
                __slots__ = ("attr_0", "attr_1")
                slots_types = dict(attr_0=int, attr_1=int)

        ok_obj = MisusedSubclass(attr_0=0, attr_1=1)  # this is ok!
        assert "attr_0" not in ok_obj.__dict__
        assert "attr_1" not in ok_obj.__dict__
        assert 0 == ok_obj.attr_0
        assert 1 == ok_obj.attr_1

        ok_obj = MisusedSubclass()
        ok_obj.update(dict(attr_0=0, attr_1=1))  # this is ok!
        assert "attr_0" not in ok_obj.__dict__
        assert "attr_1" not in ok_obj.__dict__
        assert 0 == ok_obj.attr_0
        assert 1 == ok_obj.attr_1

        bad_obj = MisusedSubclass()
        bad_obj.__dict__.update(dict(attr_0=0, attr_1=1))  # this is bad
        # See: https://github.com/rmlibre/aiootp/pull/11
        assert "attr_0" in bad_obj.__dict__
        assert "attr_1" in bad_obj.__dict__
        assert not hasattr(bad_obj, "attr_0")
        assert not hasattr(bad_obj, "attr_1")
        problem = (  # fmt: skip
            "Very strange __dict__ / __slots__ interplay didn't manifest?"
        )
        with Ignore(AttributeError, if_else=violation(problem)):
            assert bad_obj.attr_0
        with Ignore(AttributeError, if_else=violation(problem)):
            assert bad_obj.attr_1


class BaseIndexableTests(BaseVariableHoldingClassTests):
    async def test_unmapped_attributes_arent_in_dir(self) -> None:
        obj = self._type(self._items)
        if not hasattr(obj, "_UNMAPPED_ATTRIBUTES"):
            return
        if all((type(item) is str) for item in self._items):
            assert not (
                set(obj.__class__._UNMAPPED_ATTRIBUTES).difference(
                    obj.__class__._MAPPED_ATTRIBUTES
                )
            ).intersection(value for value in dir(obj))

    async def test_len_is_number_of_mapped_items_in_instance(self) -> None:
        obj = self._type(self._items)
        assert len(obj) == sum(1 for name in obj)
        assert len(obj) == len(
            set(self._items)
            .difference(getattr(obj, "_UNMAPPED_ATTRIBUTES", ()))
            .union(getattr(obj, "_MAPPED_ATTRIBUTES", ()))
        )

    async def test_indexable_iterations(self) -> None:
        obj = self._type(self._items)
        async for name in obj:
            assert name in obj
            assert name in self._items
            assert obj[name] == self._items[name]
        for name in obj:
            assert name in obj
            assert name in self._items
            assert obj[name] == self._items[name]

    async def test_item_indexing(self) -> None:
        names, values = list(self._items), list(self._items.values())
        items = {}
        all_items = {}
        obj = self._type()

        for name, value in zip(names, values):
            obj[name] = value
            all_items[name] = value
            if name in dict(obj):
                items[name] = value
                assert obj[name] == items[name]
            else:
                assert name in obj._UNMAPPED_ATTRIBUTES
            assert obj[name] == all_items[name]

        with Ignore(PermissionError, if_except=lambda _: self._frozen):
            del obj[names[0]]
            del items[names[0]]
            assert dict(obj) == dict(items)

    async def test_false_contains_logic(self) -> None:
        obj = self._type()
        assert randoms.token_bytes(4).hex() not in obj
        assert randoms.token_bits(32) not in obj
        assert None not in obj

    async def test_frozen_state_is_enforced(self) -> None:
        problem = (  # fmt: skip
            "A mutation was allowed on a frozen object."
        )
        if not self._frozen:
            return

        obj = self._type(self._items)
        with Ignore(PermissionError, if_else=violation(problem)):
            obj.__init__(self._items)

        obj = self._type(self._items)
        for name, value in self._items.items():
            with Ignore(PermissionError, if_else=violation(problem)):
                obj[name] = value


class BaseTypedSubclassDefinitionsTests(BaseVariableHoldingClassTests):
    async def test_slots_contains_all_names_in_slots_types(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "A type was created with a mismatch in variable declarations."
        )
        with Ignore(t.MissingDeclaredVariables, if_else=violation(problem)):

            class Subclass(self._type):
                __slots__ = ()
                slots_types = dict(swell=bool)

    async def test_slots_types_contains_all_names_in_slots(self) -> None:
        problem = (  # fmt: skip
            "A type was created with a mismatch in variable declarations."
        )
        with Ignore(t.MissingDeclaredVariables, if_else=violation(problem)):

            class Subclass(self._type):
                __slots__ = ("swell",)
                slots_types = {}

    async def test_declared_types_are_enforced(self) -> None:
        problem = (  # fmt: skip
            "A typed class didn't enforce a type in their `slots_types`."
        )
        obj = self._type()
        cls_set = set(obj.slots_types.values())

        for name, cls in sorted(
            obj.slots_types.items(), key=lambda _: csprng()
        ):
            wrong_cls = randoms.choice([*cls_set.difference({cls})])

            def is_vague_type(
                _: Ignore, name=name, cls=cls, wrong_cls=wrong_cls
            ) -> bool:
                error = AssertionError(
                    f"{problem=} : {name=} : {cls=} : {wrong_cls=}"
                )
                return issubclass(wrong_cls, cls) or raise_exception(error)

            with Ignore(TypeError, if_else=is_vague_type):
                obj[name] = wrong_cls()

    async def test_slots_types_correctly_allows_type_tuples(self) -> None:
        problem = (  # fmt: skip
            "A slotted value's type didn't match a type in its declared "
            "tuple of types."
        )
        assert tuples.BytesLike == (bytes, bytearray)

        class Subclass(self._type):
            __slots__ = ("bytes_like",)
            slots_types = dict(bytes_like=tuples.BytesLike)

        for acceptable_value in (b"test", bytearray(b"test")):
            ok_obj = Subclass()
            ok_obj.bytes_like = acceptable_value
            assert ok_obj.bytes_like == acceptable_value

            ok_obj = Subclass(bytes_like=acceptable_value)
            assert ok_obj.bytes_like == acceptable_value

        for unacceptable_value in ("test", list(b"test")):
            bad_obj = Subclass()
            with Ignore(TypeError, if_else=violation(problem)):
                bad_obj.bytes_like = unacceptable_value

            with Ignore(TypeError, if_else=violation(problem)):
                Subclass(bytes_like=unacceptable_value)


# Slots


class SlotsType(SlotsAttributes, Slots):
    __slots__ = tuple(BaseVariableHoldingClassTests._items)


class OpenSlotsType(SlotsAttributes, OpenSlots):
    __slots__ = tuple(BaseVariableHoldingClassTests._items)


class FrozenSlotsType(SlotsAttributes, FrozenSlots):
    __slots__ = tuple(BaseVariableHoldingClassTests._items)


class FrozenSlotsDictType(NamespaceAttributes, FrozenSlots):
    __slots__ = ("__dict__", *BaseVariableHoldingClassTests._items)


class OpenFrozenSlotsType(SlotsAttributes, OpenFrozenSlots):
    __slots__ = tuple(BaseVariableHoldingClassTests._items)


class FrozenInstanceType(SlotsAttributes, FrozenInstance):
    __slots__ = tuple(BaseVariableHoldingClassTests._items)


class TestSlots(
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseIndexableTests,
):
    _type: type = SlotsType
    _open: bool = False
    _frozen: bool = False


class TestOpenSlots(
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseIndexableTests,
):
    _type: type = OpenSlotsType
    _open: bool = True
    _frozen: bool = False


class TestFrozenSlots(
    BaseFrozenTests,
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseIndexableTests,
):
    _type: type = FrozenSlotsType
    _open: bool = False
    _frozen: bool = True


class TestFrozenSlotsDict(
    BaseFrozenTests,
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseDictLikeTests,
    BaseIndexableTests,
):
    _type: type = FrozenSlotsDictType
    _open: bool = False
    _frozen: bool = True


class TestOpenFrozenSlots(
    BaseFrozenTests,
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseIndexableTests,
):
    _type: type = OpenFrozenSlotsType
    _open: bool = True
    _frozen: bool = True


class TestFrozenInstance(
    BaseReprControlledTests,
    BaseFrozenTests,
):
    _type: type = FrozenInstanceType
    _open: bool = False
    _frozen: bool = True


# Typed Slots


class TypedSlotsType(SlotsAttributes, TypedSlots):
    __slots__ = tuple(BaseVariableHoldingClassTests._items)

    slots_types = {
        name: BaseVariableHoldingClassTests._items[name].__class__
        for name in BaseVariableHoldingClassTests._items
    }


class OpenTypedSlotsType(SlotsAttributes, OpenTypedSlots):
    __slots__ = tuple(BaseVariableHoldingClassTests._items)

    slots_types = {
        name: BaseVariableHoldingClassTests._items[name].__class__
        for name in BaseVariableHoldingClassTests._items
    }


class FrozenTypedSlotsType(SlotsAttributes, FrozenTypedSlots):
    __slots__ = tuple(BaseVariableHoldingClassTests._items)

    slots_types = {
        name: BaseVariableHoldingClassTests._items[name].__class__
        for name in BaseVariableHoldingClassTests._items
    }


class FrozenTypedSlotsDictType(NamespaceAttributes, FrozenTypedSlots):
    __slots__ = ("__dict__", *BaseVariableHoldingClassTests._items)

    slots_types = {
        name: BaseVariableHoldingClassTests._items[name].__class__
        for name in BaseVariableHoldingClassTests._items
    }


class OpenFrozenTypedSlotsType(SlotsAttributes, OpenFrozenTypedSlots):
    __slots__ = tuple(BaseVariableHoldingClassTests._items)

    slots_types = {
        name: BaseVariableHoldingClassTests._items[name].__class__
        for name in BaseVariableHoldingClassTests._items
    }


class TestTypedSlots(
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseIndexableTests,
    BaseTypedSubclassDefinitionsTests,
):
    _type: type = TypedSlotsType
    _open: bool = False
    _frozen: bool = False


class TestOpenTypedSlots(
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseIndexableTests,
    BaseTypedSubclassDefinitionsTests,
):
    _type: type = OpenTypedSlotsType
    _open: bool = True
    _frozen: bool = False


class TestFrozenTypedSlots(
    BaseFrozenTests,
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseIndexableTests,
    BaseTypedSubclassDefinitionsTests,
):
    _type: type = FrozenTypedSlotsType
    _open: bool = False
    _frozen: bool = True


class TestFrozenTypedSlotsDict(
    BaseFrozenTests,
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseDictLikeTests,
    BaseIndexableTests,
    BaseTypedSubclassDefinitionsTests,
):
    _type: type = FrozenTypedSlotsDictType
    _open: bool = False
    _frozen: bool = True


class TestOpenFrozenTypedSlots(
    BaseFrozenTests,
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseIndexableTests,
    BaseTypedSubclassDefinitionsTests,
):
    _type: type = OpenFrozenTypedSlotsType
    _open: bool = True
    _frozen: bool = True


# Configs


class ConfigType(SlotsAttributes, Config):
    __slots__ = tuple(BaseVariableHoldingClassTests._items)

    slots_types = {
        name: BaseVariableHoldingClassTests._items[name].__class__
        for name in BaseVariableHoldingClassTests._items
    }


class TestConfigType(
    BaseFrozenTests,
    BaseReprControlledTests,
    BaseIndexableTests,
    BaseTypedSubclassDefinitionsTests,
):
    _type: type = ConfigType
    _open: bool = True
    _frozen: bool = True


# Namespaces


class NamespaceType(NamespaceAttributes, Namespace):
    pass


class OpenNamespaceType(NamespaceAttributes, OpenNamespace):
    pass


class FrozenNamespaceType(NamespaceAttributes, FrozenNamespace):
    pass


class OpenFrozenNamespaceType(NamespaceAttributes, OpenFrozenNamespace):
    pass


class TestNamespaceMapping(
    BaseDictLikeTests,
    BaseIndexableTests,
):
    _type: type = t.NamespaceMapping
    _open: bool = True
    _frozen: bool = False


class TestNamespace(
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseDictLikeTests,
    BaseIndexableTests,
):
    _type: type = NamespaceType
    _open: bool = False
    _frozen: bool = False


class TestOpenNamespace(
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseDictLikeTests,
    BaseIndexableTests,
):
    _type: type = OpenNamespaceType
    _open: bool = True
    _frozen: bool = False


class TestFrozenNamespace(
    BaseFrozenTests,
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseDictLikeTests,
    BaseIndexableTests,
):
    _type: type = FrozenNamespaceType
    _open: bool = False
    _frozen: bool = True


class TestOpenFrozenNamespace(
    BaseFrozenTests,
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseDictLikeTests,
    BaseIndexableTests,
):
    _type: type = OpenFrozenNamespaceType
    _open: bool = True
    _frozen: bool = True


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Streaming manager classes for the package's ciphers.
"""

__all__ = ["AsyncCipherStream", "CipherStream"]


import io
from collections import deque
from hmac import compare_digest
from secrets import token_bytes

from aiootp._typing import Typing as t
from aiootp._constants.misc import DEFAULT_AAD
from aiootp._exceptions import CipherStreamIsClosed
from aiootp._gentools import apopleft, popleft, abatch, batch
from aiootp.asynchs import AsyncInit, ConcurrencyGuard, asleep

from .cipher_stream_properties import CipherStreamProperties


class AsyncCipherStream(CipherStreamProperties, metaclass=AsyncInit):
    """
    A high-level public interface to the package's online, salt reuse /
    misuse resistant, tweakable, AEAD ciphers.

    Validation using `StreamHMAC` `block_id`'s provides robustness to
    communication channels. It does so by detecting forgeries & out-of-
    order messages without changing its internal state. If a tag fails
    to validate, the cipher stream can continue functioning once the
    authentic messages, in their correct order, begin coming in.
    --------
    HOWEVER: The more blocks that are allowed to fail without aborting,
    -------- the more chances an attacker has to spoof the auth tag.

     _____________________________________
    |                                     |
    |      Usage Example: Encryption      |
    |_____________________________________|


    stream = await AsyncCipherStream(key, aad=session.transcript)
    session.transmit(salt=stream.salt, iv=stream.iv)

    for plaintext in session.upload.buffer(4 * stream.PACKETSIZE):
        await stream.abuffer(plaintext)
        async for block_id, ciphertext in stream:
            session.send_packet(block_id + ciphertext)

    async for block_id, ciphertext in stream.afinalize():
        session.send_packet(block_id + ciphertext)

    # Send the stream termination / authentication tag. <--------
    session.transmit(shmac=await stream.shmac.result)
    """

    __slots__ = (
        "_buffer",
        "_byte_count",
        "_config",
        "_digesting_now",
        "_finalizing_now",
        "_key_bundle",
        "_padding",
        "_stream",
        "shmac",
    )

    _MAX_SIMULTANEOUS_BUFFERS: int = 1024

    async def __init__(
        self,
        cipher: t.CipherInterfaceType,
        *,
        salt: t.Optional[bytes] = None,
        aad: bytes = DEFAULT_AAD,
    ) -> None:
        """
        Derives encryption keys & initializes a mutable buffer to
        automatically prepare plaintext given by the user with the
        necessary padding.

        `salt`: A [pseudo]random salt that may be supplied by the user. By
                default it's sent in the clear attached to the ciphertext.
                Thus it may simplify implementing efficient features, such
                as search or routing, though care must still be taken when
                considering how leaking such metadata may be harmful.

                Keeping this value constant is strongly discouraged. Though,
                the cipher's salt misuse-reuse resistance is ruled by the
                combination of the automatically incorporated `timestamp`,
                `iv`, & `siv_key`. The risk calculation starts with setting
                r = len(iv + siv_key) / 3. Then, all else staying constant,
                once 256**r messages are encrypted within a second, each
                additional encrypted message within that same second begins
                to have more than a 256**(-r) chance of generating a repeat
                context.

                See: https://github.com/rmlibre/aiootp/issues/16

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.
        """
        self._config = cipher._config
        self._padding = cipher._padding
        self._byte_count = 0
        self._digesting_now = deque(maxlen=self._MAX_SIMULTANEOUS_BUFFERS)
        self._finalizing_now = deque()  # don't let maxlen remove entries
        self._buffer = buffer = deque([self._padding.start_padding()])
        self._key_bundle = key_bundle = await cipher._KeyAADBundle(
            kdfs=cipher._kdfs, salt=salt, aad=aad
        ).async_mode()
        self.shmac = cipher._StreamHMAC(key_bundle)._for_encryption()
        self._stream = cipher._Junction.abytes_encipher(
            apopleft(buffer), shmac=self.shmac
        )

    @property
    def _iter_shortcuts(
        self,
    ) -> t.Tuple[
        t.Callable[..., bytes],
        t.Deque[bytes],
        t.Callable[[None], bytes],
    ]:
        """
        Returns method pointers so calls in tight loops during
        processing don't have to continually getattr on instance
        objects.
        """
        return self.shmac.anext_block_id, self._buffer, self._stream.asend

    @property
    def _buffer_shortcuts(
        self,
    ) -> t.Tuple[t.Deque[bytes], t.Callable[[bytes], None]]:
        """
        Returns method pointers so calls in tight loops during
        processing don't have to continually getattr on instance
        objects.
        """
        return self._buffer, self._buffer.append

    async def __aiter__(
        self,
    ) -> t.AsyncGenerator[t.Tuple[bytes, bytes], None]:
        """
        Allows the object to be entered in for-loops an unlimited amount
        of times in the process of gathering data to buffer for the
        stream. This leads to a very pythonic API.
        """
        MIN_STREAM_QUEUE = self._config.MIN_STREAM_QUEUE
        anext_block_id, buffer, cipher = self._iter_shortcuts
        while len(buffer) > MIN_STREAM_QUEUE:
            block = await cipher(None)
            yield await anext_block_id(block), block

    async def afinalize(
        self,
    ) -> t.AsyncGenerator[t.Tuple[bytes, bytes], None]:
        """
        Instructs the instance to finish receiving data into the buffer
        & to flush all ciphertext results out to the user.

         _____________________________________
        |                                     |
        |      Usage Example: Encryption      |
        |_____________________________________|


        for plaintext in session.upload.buffer(4 * stream.PACKETSIZE):
            await stream.abuffer(plaintext)
            async for block_id, ciphertext in stream:
                session.send_packet(block_id + ciphertext)

        async for block_id, ciphertext in stream.afinalize():  # <------
            session.send_packet(block_id + ciphertext)
        """
        self._finalizing_now.append(token := token_bytes(32))
        if not compare_digest(token, self._finalizing_now[0]):
            raise ConcurrencyGuard.IncoherentConcurrencyState

        async with ConcurrencyGuard(self._digesting_now, token=token):
            end_padding = await self._padding.aend_padding(self._byte_count)
            final_blocks = abatch(
                self._buffer.pop() + end_padding,
                size=self._config.BLOCKSIZE,
            )
            async for block in final_blocks:
                self._buffer.append(block)
            while self._buffer:
                block = await self._stream.asend(None)
                block_id = await self.shmac.anext_block_id(block)
                yield block_id, block
            await self.shmac.afinalize()

    async def _adigest_data(
        self,
        data: t.Callable[[int], bytes],
        buffer: t.Deque[bytes],
        append: t.Callable[[bytes], None],
    ) -> None:
        """
        Prepares the input plaintext `data` for encryption by dividing
        it into blocksize chunks.
        """
        BLOCKSIZE = self._config.BLOCKSIZE
        if buffer and len(buffer[-1]) != BLOCKSIZE:
            missing_bytes = BLOCKSIZE - len(buffer[-1])
            chunk = data(missing_bytes)
            buffer[-1] += chunk
            if len(chunk) != missing_bytes:
                return
        while True:
            await asleep()
            block = data(BLOCKSIZE)
            append(block)
            if len(block) != BLOCKSIZE:
                break

    async def abuffer(self, data: bytes) -> t.Self:
        """
        Prepares the input plaintext `data` for encryption by dividing
        it into blocksize chunks & taking plaintext measuremenets for
        automated message padding.

         _____________________________________
        |                                     |
        |      Usage Example: Encryption      |
        |_____________________________________|


        for plaintext in session.upload.buffer(4 * stream.PACKETSIZE):
            await stream.abuffer(plaintext)  # <------------------------
            async for block_id, ciphertext in stream:
                session.send_packet(block_id + ciphertext)

        async for block_id, ciphertext in stream.afinalize():
            session.send_packet(block_id + ciphertext)
        """
        async with ConcurrencyGuard(self._digesting_now):
            if await self._aconstant_time_final_context_is_done():
                raise CipherStreamIsClosed
            self._byte_count += len(data)
            data = io.BytesIO(data).read
            _buffer, append = self._buffer_shortcuts
            await self._adigest_data(data, _buffer, append)
        return self


class CipherStream(CipherStreamProperties):
    """
    A high-level public interface to the package's online, salt reuse /
    misuse resistant, tweakable, AEAD ciphers.

    Validation using `StreamHMAC` `block_id`'s provides robustness to
    communication channels. It does so by detecting forgeries & out-of-
    order messages without changing its internal state. If a tag fails
    to validate, the cipher stream can continue functioning once the
    authentic messages, in their correct order, begin coming in.
    --------
    HOWEVER: The more blocks that are allowed to fail without aborting,
    -------- the more chances an attacker has to spoof the auth tag.

     _____________________________________
    |                                     |
    |      Usage Example: Encryption      |
    |_____________________________________|


    stream = CipherStream(key, aad=session.transcript)
    session.transmit(salt=stream.salt, iv=stream.iv)

    for plaintext in session.upload.buffer(4 * stream.PACKETSIZE):
        stream.buffer(plaintext)
        for block_id, ciphertext in stream:
            session.send_packet(block_id + ciphertext)

    for block_id, ciphertext in stream.finalize():
        session.send_packet(block_id + ciphertext)

    # Send the stream termination / authentication tag. <--------
    session.transmit(shmac=stream.shmac.result)
    """

    __slots__ = (
        "_buffer",
        "_byte_count",
        "_config",
        "_digesting_now",
        "_finalizing_now",
        "_key_bundle",
        "_padding",
        "_stream",
        "shmac",
    )

    _MAX_SIMULTANEOUS_BUFFERS: int = 1024

    def __init__(
        self,
        cipher: t.CipherInterfaceType,
        *,
        salt: t.Optional[bytes] = None,
        aad: bytes = DEFAULT_AAD,
    ) -> None:
        """
        Derives encryption keys & initializes a mutable buffer to
        automatically prepare plaintext given by the user with the
        necessary padding.

        `salt`: A [pseudo]random salt that may be supplied by the user. By
                default it's sent in the clear attached to the ciphertext.
                Thus it may simplify implementing efficient features, such
                as search or routing, though care must still be taken when
                considering how leaking such metadata may be harmful.

                Keeping this value constant is strongly discouraged. Though,
                the cipher's salt misuse-reuse resistance is ruled by the
                combination of the automatically incorporated `timestamp`,
                `iv`, & `siv_key`. The risk calculation starts with setting
                r = len(iv + siv_key) / 3. Then, all else staying constant,
                once 256**r messages are encrypted within a second, each
                additional encrypted message within that same second begins
                to have more than a 256**(-r) chance of generating a repeat
                context.

                See: https://github.com/rmlibre/aiootp/issues/16

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.
        """
        self._config = cipher._config
        self._padding = cipher._padding
        self._byte_count = 0
        self._digesting_now = deque(maxlen=self._MAX_SIMULTANEOUS_BUFFERS)
        self._finalizing_now = deque()  # don't let maxlen remove entries
        self._buffer = buffer = deque([self._padding.start_padding()])
        self._key_bundle = key_bundle = cipher._KeyAADBundle(
            kdfs=cipher._kdfs, salt=salt, aad=aad
        ).sync_mode()
        self.shmac = cipher._StreamHMAC(key_bundle)._for_encryption()
        self._stream = cipher._Junction.bytes_encipher(
            popleft(buffer), shmac=self.shmac
        )

    @property
    def _iter_shortcuts(
        self,
    ) -> t.Tuple[
        t.Callable[..., bytes],
        t.Deque[bytes],
        t.Callable[[None], bytes],
    ]:
        """
        Returns method pointers so calls in tight loops during
        processing don't have to continually getattr on instance
        objects.
        """
        return self.shmac.next_block_id, self._buffer, self._stream.send

    @property
    def _buffer_shortcuts(
        self,
    ) -> t.Tuple[t.Deque[bytes], t.Callable[[bytes], None]]:
        """
        Returns method pointers so calls in tight loops during
        processing don't have to continually getattr on instance
        objects.
        """
        return self._buffer, self._buffer.append

    def __iter__(self) -> t.Generator[t.Tuple[bytes, bytes], None, None]:
        """
        Allows the object to be entered in for-loops an unlimited amount
        of times in the process of gathering data to buffer for the
        stream. This leads to a very pythonic API.
        """
        MIN_STREAM_QUEUE = self._config.MIN_STREAM_QUEUE
        next_block_id, buffer, cipher = self._iter_shortcuts
        while len(buffer) > MIN_STREAM_QUEUE:
            block = cipher(None)
            yield next_block_id(block), block

    def finalize(self) -> t.Generator[t.Tuple[bytes, bytes], None, None]:
        """
        Instructs the instance to finish receiving data into the buffer
        & to flush all ciphertext results out to the user.

         _____________________________________
        |                                     |
        |      Usage Example: Encryption      |
        |_____________________________________|


        for plaintext in session.upload.buffer(4 * stream.PACKETSIZE):
            stream.buffer(plaintext)
            for block_id, ciphertext in stream:
                session.send_packet(block_id + ciphertext)

        for block_id, ciphertext in stream.finalize():  # <-------------
            session.send_packet(block_id + ciphertext)
        """
        self._finalizing_now.append(token := token_bytes(32))
        if not compare_digest(token, self._finalizing_now[0]):
            raise ConcurrencyGuard.IncoherentConcurrencyState

        with ConcurrencyGuard(
            self._digesting_now, probe_delay=0.0001, token=token
        ):
            end_padding = self._padding.end_padding(self._byte_count)
            final_blocks = batch(
                self._buffer.pop() + end_padding,
                size=self._config.BLOCKSIZE,
            )
            for block in final_blocks:
                self._buffer.append(block)
            while self._buffer:
                block = self._stream.send(None)
                block_id = self.shmac.next_block_id(block)
                yield block_id, block
            self.shmac.finalize()

    def _digest_data(
        self,
        data: t.Callable[[int], bytes],
        buffer: t.Deque[bytes],
        append: t.Callable[[bytes], None],
    ) -> None:
        """
        Prepares the input plaintext `data` for encryption by dividing
        it into blocksize chunks.
        """
        BLOCKSIZE = self._config.BLOCKSIZE
        if buffer and len(buffer[-1]) != BLOCKSIZE:
            missing_bytes = BLOCKSIZE - len(buffer[-1])
            chunk = data(missing_bytes)
            buffer[-1] += chunk
            if len(chunk) != missing_bytes:
                return
        while True:
            block = data(BLOCKSIZE)
            append(block)
            if len(block) != BLOCKSIZE:
                break

    def buffer(self, data: bytes) -> t.Self:
        """
        Prepares the input plaintext `data` for encryption by dividing
        it into blocksize chunks & taking plaintext measuremenets for
        automated message padding.

         _____________________________________
        |                                     |
        |      Usage Example: Encryption      |
        |_____________________________________|


        for plaintext in session.upload.buffer(4 * stream.PACKETSIZE):
            stream.buffer(plaintext)  # <-------------------------------
            for block_id, ciphertext in stream:
                session.send_packet(block_id + ciphertext)

        for block_id, ciphertext in stream.finalize():
            session.send_packet(block_id + ciphertext)
        """
        with ConcurrencyGuard(self._digesting_now, probe_delay=0.0001):
            if self._constant_time_final_context_is_done():
                raise CipherStreamIsClosed
            self._byte_count += len(data)
            data = io.BytesIO(data).read
            _buffer, append = self._buffer_shortcuts
            self._digest_data(data, _buffer, append)
        return self


module_api = dict(
    AsyncCipherStream=t.add_type(AsyncCipherStream),
    CipherStream=t.add_type(CipherStream),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A general type defining an interface to Curve25519 operations.
"""

__all__ = ["Base25519"]


from cryptography.hazmat.primitives.asymmetric.x25519 import (
    X25519PublicKey,
    X25519PrivateKey,
)
from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PublicKey,
    Ed25519PrivateKey,
)

from aiootp._typing import Typing as t
from aiootp._exceptions import Issue
from aiootp.asynchs import asleep
from aiootp.commons import FrozenInstance


class Base25519(FrozenInstance):
    """
    Collects the shared functionality between the `X25519` & `Ed25519`
    classes.
    """

    __slots__ = ("_public_key", "_secret_key")

    PublicKey: type = None
    SecretKey: type = None

    def __init__(self, /) -> None:
        self._secret_key = None
        self._public_key = None

    @classmethod
    def _process_secret_key(
        cls, secret_key: t.Union[bytes, t.SecretKeyType], /
    ) -> t.SecretKeyType:
        """
        If `secret_key` is either of `bytes` or `cls.SecretKey` type,
        returns an instance of `cls.SecretKey`. Otherwise raises
        `TypeError`.
        """
        key_type = secret_key.__class__
        if key_type is bytes:
            return cls.SecretKey.from_private_bytes(secret_key)
        elif issubclass(key_type, cls.SecretKey):
            return secret_key
        else:
            raise Issue.value_must_be_type("secret_key", "valid key type")

    @classmethod
    def _process_public_key(
        cls, public_key: t.Union[bytes, t.PublicKeyType], /
    ) -> t.PublicKeyType:
        """
        If `public_key` is either of `bytes` or `cls.PublicKey` type,
        returns an instance of `cls.PublicKey`. Otherwise raises
        `TypeError`.
        """
        key_type = public_key.__class__
        if key_type is bytes:
            return cls.PublicKey.from_public_bytes(public_key)
        elif issubclass(key_type, cls.PublicKey):
            return public_key
        else:
            raise Issue.value_must_be_type("public_key", "valid key type")

    async def aimport_secret_key(
        self, secret_key: t.Union[bytes, t.SecretKeyType], /
    ) -> t.Self:
        """
        Populates an instance from the received `secret_key` that is
        of either `bytes` or `self.SecretKey` type.
        """
        await asleep()
        return self.import_secret_key(secret_key)

    def import_secret_key(
        self, secret_key: t.Union[bytes, t.SecretKeyType], /
    ) -> t.Self:
        """
        Populates an instance from the received `secret_key` that is
        of either `bytes` or `self.SecretKey` type.
        """
        if self.has_public_key():
            raise Issue.cant_reassign_attribute("public_key")
        insert = object.__setattr__
        insert(self, "_secret_key", self._process_secret_key(secret_key))
        insert(self, "_public_key", self._secret_key.public_key())
        return self

    async def aimport_public_key(
        self, public_key: t.Union[bytes, t.PublicKeyType], /
    ) -> t.Self:
        """
        Populates an instance from the received `public_key` that is
        of either `bytes` or `self.PublicKey` type.
        """
        await asleep()
        return self.import_public_key(public_key)

    def import_public_key(
        self, public_key: t.Union[bytes, t.PublicKeyType], /
    ) -> t.Self:
        """
        Populates an instance from the received `public_key` that is
        of either `bytes` or `self.PublicKey` type.
        """
        if self.has_public_key():
            raise Issue.cant_reassign_attribute("public_key")
        insert = object.__setattr__
        insert(self, "_secret_key", None)
        insert(self, "_public_key", self._process_public_key(public_key))
        return self

    async def agenerate(self, /) -> t.Self:
        """
        Populates the instance with a newly generated secret key.
        """
        return await self.aimport_secret_key(self.SecretKey.generate())

    def generate(self, /) -> t.Self:
        """
        Populates the instance with a newly generated secret key.
        """
        return self.import_secret_key(self.SecretKey.generate())

    @property
    def secret_key(self, /) -> t.Optional[t.SecretKeyType]:
        """
        Returns the instance's secret key object.
        """
        return self._secret_key

    @property
    def public_key(self, /) -> t.Optional[t.PublicKeyType]:
        """
        Returns the instance's public key object.
        """
        return self._public_key

    @property
    def secret_bytes(self, /) -> bytes:
        """
        Returns the secret bytes of the instance's secret key.
        """
        return self._secret_key.private_bytes_raw()

    @property
    def public_bytes(self, /) -> bytes:
        """
        Returns the public bytes of the instance's public key.
        """
        return self._public_key.public_bytes_raw()

    def has_secret_key(self, /) -> bool:
        """
        Returns a boolean of whether the instance contains a secret key.
        """
        return self._secret_key is not None

    def has_public_key(self, /) -> bool:
        """
        Returns a boolean of whether the instance contains a public key.
        """
        return self._public_key is not None


module_api = dict(
    Base25519=t.add_type(Base25519),
    Ed25519PublicKey=t.add_type(Ed25519PublicKey),
    Ed25519PrivateKey=t.add_type(Ed25519PrivateKey),
    X25519PublicKey=t.add_type(X25519PublicKey),
    X25519PrivateKey=t.add_type(X25519PrivateKey),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A type for `(Async)Database` domain-separation tasks.
"""

__all__ = ["DBDomains"]


from aiootp._typing import Typing as t
from aiootp.generics import DomainEncoder


class DBDomains(DomainEncoder):
    """
    A container for database-specific domain constants.
    """

    __slots__ = ()

    _encode: t.Callable = lambda constant: DomainEncoder.encode_constant(
        constant, domain=b"database_constants", size=16
    )

    ROOT_KDF: bytes = _encode("root_kdf")
    ROOT_FILENAME: bytes = _encode("root_filename")
    ROOT_SALT: bytes = _encode("root_salt")
    METATAG: bytes = _encode("metatag")
    HMAC: bytes = _encode("hmac")
    GIST: bytes = _encode("profile_credential_gist")
    TMP_PREKEY: bytes = _encode("temporary_profile_prekey")
    PROFILE_LOGIN_KEY: bytes = _encode("profile_login_key")
    MANIFEST: bytes = _encode("manifest")
    FILENAME: bytes = _encode("filename")
    METATAG_KEY: bytes = _encode("metatag_key")
    DEVICE_SALT: bytes = _encode(b"device_salt")
    CIPHER: bytes = _encode("cipher")


module_api = dict(
    DBDomains=t.add_type(DBDomains),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)
